#!/usr/bin/env python3
import warnings
from collections import namedtuple
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    Generic,
    List,
    NamedTuple,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from captum._utils.common import (
    _expand_additional_forward_args,
    _format_additional_forward_args,
    _reduce_list,
)
from captum.attr import Max, Mean, Min, Summarizer
from captum.log import log_usage
from captum.robust._core.perturbation import Perturbation
from torch import Tensor

ORIGINAL_KEY = "Original"

MetricResultType = TypeVar(
    "MetricResultType", float, Tensor, Tuple[Union[float, Tensor], ...]
)


class AttackInfo(NamedTuple):
    attack_fn: Union[Perturbation, Callable]
    name: str
    num_attempts: int
    apply_before_preproc: bool
    attack_kwargs: Dict[str, Any]
    additional_args: List[str]


def agg_metric(inp):
    if isinstance(inp, Tensor):
        return inp.mean(dim=0)
    elif isinstance(inp, tuple):
        return tuple(agg_metric(elem) for elem in inp)
    return inp


class AttackComparator(Generic[MetricResultType]):
    r"""
    Allows measuring model robustness for a given attack or set of attacks. This class
    can be used with any metric(s) as well as any set of attacks, either based on
    attacks / perturbations from captum.robust such as FGSM or PGD or external
    augmentation methods or perturbations such as torchvision transforms.
    """

    def __init__(
        self,
        forward_func: Callable,
        metric: Callable[..., MetricResultType],
        preproc_fn: Optional[Callable] = None,
    ) -> None:
        r"""
        Args:
            forward_func (Callable or torch.nn.Module): This can either be an instance
                of pytorch model or any modification of a model's forward
                function.

            metric (Callable): This function is applied to the model output in
                order to compute the desired performance metric or metrics.
                This function should have the following signature::

                    >>> def model_metric(model_out: Tensor, **kwargs: Any)
                    >>>     -> Union[float, Tensor, Tuple[Union[float, Tensor], ...]:

                All kwargs provided to evaluate are provided to the metric function,
                following the model output. A single metric can be returned as
                a float or tensor, and multiple metrics should be returned as either
                a tuple or named tuple of floats or tensors. For a tensor metric,
                the first dimension should match the batch size, corresponding to
                metrics for each example. Tensor metrics are averaged over the first
                dimension when aggregating multiple batch results.
                If tensor metrics represent results for the full batch, the size of the
                first dimension should be 1.

            preproc_fn (Callable, optional): Optional method applied to inputs. Output
                of preproc_fn is then provided as input to model, in addition to
                additional_forward_args provided to evaluate.
                Default: ``None``
        """
        self.forward_func = forward_func
        self.metric: Callable = metric
        self.preproc_fn = preproc_fn
        self.attacks: Dict[str, AttackInfo] = {}
        self.summary_results: Dict[str, Summarizer] = {}
        self.metric_aggregator = agg_metric
        self.batch_stats = [Mean, Min, Max]
        self.aggregate_stats = [Mean]
        self.summary_results = {}
        self.out_format = None

    def add_attack(
        self,
        attack: Union[Perturbation, Callable],
        name: Optional[str] = None,
        num_attempts: int = 1,
        apply_before_preproc: bool = True,
        attack_kwargs: Optional[Dict[str, Any]] = None,
        additional_attack_arg_names: Optional[List[str]] = None,
    ) -> None:
        r"""
        Adds attack to be evaluated when calling evaluate.

        Args:

            attack (Perturbation or Callable): This can either be an instance
                of a Captum Perturbation / Attack
                or any other perturbation or attack function such
                as a torchvision transform.

            name (str, optional): Name or identifier for attack, used as key for
                attack results. This defaults to attack.__class__.__name__
                if not provided and must be unique for all added attacks.
                Default: ``None``

            num_attempts (int, optional): Number of attempts that attack should be
                repeated. This should only be set to > 1 for non-deterministic
                attacks. The minimum, maximum, and average (best, worst, and
                average case) are tracked for attack attempts.
                Default: ``1``

            apply_before_preproc (bool, optional): Defines whether attack should be
                applied before or after preproc function.
                Default: ``True``

            attack_kwargs (dict, optional): Additional arguments to be provided to
                given attack. This should be provided as a dictionary of keyword
                arguments.
                Default: ``None``

            additional_attack_arg_names (list[str], optional): Any additional
                arguments for the attack which are specific to the particular input
                example or batch. An example of this is target, which is necessary
                for some attacks such as FGSM or PGD. These arguments are included
                if provided as a kwarg to evaluate.
                Default: ``None``
        """
        if name is None:
            name = attack.__class__.__name__

        if attack_kwargs is None:
            attack_kwargs = {}

        if additional_attack_arg_names is None:
            additional_attack_arg_names = []

        if name in self.attacks:
            raise RuntimeError(
                "Cannot add attack with same name as existing attack {}".format(name)
            )

        self.attacks[name] = AttackInfo(
            attack_fn=attack,
            name=name,
            num_attempts=num_attempts,
            apply_before_preproc=apply_before_preproc,
            attack_kwargs=attack_kwargs,
            additional_args=additional_attack_arg_names,
        )

    def _format_summary(
        self, summary: Union[Dict, List[Dict]]
    ) -> Dict[str, MetricResultType]:
        r"""
        This method reformats a given summary; particularly for tuples,
        the Summarizer's summary format is a list of dictionaries,
        each containing the summary for the corresponding elements.
        We reformat this to return a dictionary with tuples containing
        the summary results.
        """
        if isinstance(summary, dict):
            return summary
        else:
            summary_dict: Dict[str, Tuple] = {}
            for key in summary[0]:
                summary_dict[key] = tuple(s[key] for s in summary)
                if self.out_format:
                    summary_dict[key] = self.out_format(*summary_dict[key])
            return summary_dict  # type: ignore

    def _update_out_format(
        self, out_metric: Union[float, Tensor, Tuple[Union[float, Tensor], ...]]
    ) -> None:
        if (
            not self.out_format
            and isinstance(out_metric, tuple)
            and hasattr(out_metric, "_fields")
        ):
            self.out_format = namedtuple(  # type: ignore
                type(out_metric).__name__, cast(NamedTuple, out_metric)._fields
            )

    def _evaluate_batch(
        self,
        input_list: List[Any],
        additional_forward_args: Optional[Tuple],
        key_list: List[str],
        batch_summarizers: Dict[str, Summarizer],
        metric_kwargs: Dict[str, Any],
    ) -> None:
        if additional_forward_args is None:
            additional_forward_args = ()
        if len(input_list) == 1:
            model_out = self.forward_func(input_list[0], *additional_forward_args)
            out_metric = self.metric(model_out, **metric_kwargs)
            self._update_out_format(out_metric)
            batch_summarizers[key_list[0]].update(out_metric)
        else:
            batched_inps = _reduce_list(input_list)
            model_out = self.forward_func(batched_inps, *additional_forward_args)
            current_count = 0
            for i in range(len(input_list)):
                batch_size = (
                    input_list[i].shape[0]
                    if isinstance(input_list[i], Tensor)
                    else input_list[i][0].shape[0]
                )
                out_metric = self.metric(
                    model_out[current_count : current_count + batch_size],
                    **metric_kwargs,
                )
                self._update_out_format(out_metric)
                batch_summarizers[key_list[i]].update(out_metric)
                current_count += batch_size

    @log_usage()
    def evaluate(
        self,
        inputs: Any,
        additional_forward_args: Any = None,
        perturbations_per_eval: int = 1,
        **kwargs,
    ) -> Dict[str, Union[MetricResultType, Dict[str, MetricResultType]]]:
        r"""
        Evaluate model and attack performance on provided inputs

        Args:

            inputs (Any): Input for which attack metrics
                are computed. It can be provided as a tensor, tuple of tensors,
                or any raw input type (e.g. PIL image or text string).
                This input is provided directly as input to preproc function as well
                as any attack applied before preprocessing. If no pre-processing
                function is provided, this input is provided directly to the main
                model and all attacks.

            additional_forward_args (Any, optional): If the forward function
                requires additional arguments other than the preprocessing
                outputs (or inputs if preproc_fn is None), this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a
                tuple containing multiple additional arguments including
                tensors or any arbitrary python types. These arguments
                are provided to forward_func in order following the
                arguments in inputs.
                For a tensor, the first dimension of the tensor must
                correspond to the number of examples. For all other types,
                the given argument is used for all forward evaluations.
                Default: ``None``
            perturbations_per_eval (int, optional): Allows perturbations of multiple
                attacks to be grouped and evaluated in one call of forward_fn
                Each forward pass will contain a maximum of
                perturbations_per_eval * #examples samples.
                For DataParallel models, each batch is split among the
                available devices, so evaluations on each available
                device contain at most
                (perturbations_per_eval * #examples) / num_devices
                samples.
                In order to apply this functionality, the output of preproc_fn
                (or inputs itself if no preproc_fn is provided) must be a tensor
                or tuple of tensors.
                Default: ``1``
            kwargs (Any, optional): Additional keyword arguments provided to metric
                function as well as selected attacks based on chosen additional_args.
                Default: ``None``

        Returns:

        - **attack results** Dict: str -> Dict[str, Union[Tensor, Tuple[Tensor, ...]]]:
                Dictionary containing attack results for provided batch.
                Maps attack name to dictionary,
                containing best-case, worst-case and average-case results for attack.
                Dictionary contains keys "mean", "max" and "min" when num_attempts > 1
                and only "mean" for num_attempts = 1, which contains the (single) metric
                result for the attack attempt.
                An additional key of 'Original' is included with metric results
                without any perturbations.


        Examples::

        >>> def accuracy_metric(model_out: Tensor, targets: Tensor):
        >>>     return torch.argmax(model_out, dim=1) == targets).float()

        >>> attack_metric = AttackComparator(model=resnet18,
                                             metric=accuracy_metric,
                                             preproc_fn=normalize)

        >>> random_rotation = transforms.RandomRotation()
        >>> jitter = transforms.ColorJitter()

        >>> attack_metric.add_attack(random_rotation, "Random Rotation",
        >>>                          num_attempts = 5)
        >>> attack_metric.add_attack((jitter, "Jitter", num_attempts = 1)
        >>> attack_metric.add_attack(FGSM(resnet18), "FGSM 0.1", num_attempts = 1,
        >>>                          apply_before_preproc=False,
        >>>                          attack_kwargs={epsilon: 0.1},
        >>>                          additional_args=["targets"])

        >>> for images, labels in dataloader:
        >>>     batch_results = attack_metric.evaluate(inputs=images, targets=labels)
        """
        additional_forward_args = _format_additional_forward_args(
            additional_forward_args
        )
        expanded_additional_args = (
            _expand_additional_forward_args(
                additional_forward_args, perturbations_per_eval
            )
            if perturbations_per_eval > 1
            else additional_forward_args
        )

        preproc_input = None
        if self.preproc_fn is not None:
            preproc_input = self.preproc_fn(inputs)
        else:
            preproc_input = inputs

        input_list = [preproc_input]
        key_list = [ORIGINAL_KEY]

        batch_summarizers = {ORIGINAL_KEY: Summarizer([Mean()])}
        if ORIGINAL_KEY not in self.summary_results:
            self.summary_results[ORIGINAL_KEY] = Summarizer(
                [stat() for stat in self.aggregate_stats]
            )

        def _check_and_evaluate(input_list, key_list):
            if len(input_list) == perturbations_per_eval:
                self._evaluate_batch(
                    input_list,
                    expanded_additional_args,
                    key_list,
                    batch_summarizers,
                    kwargs,
                )
                return [], []
            return input_list, key_list

        input_list, key_list = _check_and_evaluate(input_list, key_list)

        for attack_key in self.attacks:
            attack = self.attacks[attack_key]
            if attack.num_attempts > 1:
                stats = [stat() for stat in self.batch_stats]
            else:
                stats = [Mean()]
            batch_summarizers[attack.name] = Summarizer(stats)
            additional_attack_args = {}
            for key in attack.additional_args:
                if key not in kwargs:
                    warnings.warn(
                        f"Additional sample arg {key} not provided for {attack_key}"
                    )
                else:
                    additional_attack_args[key] = kwargs[key]

            for _ in range(attack.num_attempts):
                if attack.apply_before_preproc:
                    attacked_inp = attack.attack_fn(
                        inputs, **additional_attack_args, **attack.attack_kwargs
                    )
                    preproc_attacked_inp = (
                        self.preproc_fn(attacked_inp)
                        if self.preproc_fn
                        else attacked_inp
                    )
                else:
                    preproc_attacked_inp = attack.attack_fn(
                        preproc_input, **additional_attack_args, **attack.attack_kwargs
                    )

                input_list.append(preproc_attacked_inp)
                key_list.append(attack.name)

                input_list, key_list = _check_and_evaluate(input_list, key_list)

        if len(input_list) > 0:
            final_add_args = _expand_additional_forward_args(
                additional_forward_args, len(input_list)
            )
            self._evaluate_batch(
                input_list, final_add_args, key_list, batch_summarizers, kwargs
            )

        return self._parse_and_update_results(batch_summarizers)

    def _parse_and_update_results(
        self, batch_summarizers: Dict[str, Summarizer]
    ) -> Dict[str, Union[MetricResultType, Dict[str, MetricResultType]]]:
        results: Dict[str, Union[MetricResultType, Dict[str, MetricResultType]]] = {
            ORIGINAL_KEY: self._format_summary(
                cast(Union[Dict, List], batch_summarizers[ORIGINAL_KEY].summary)
            )["mean"]
        }
        self.summary_results[ORIGINAL_KEY].update(
            self.metric_aggregator(results[ORIGINAL_KEY])
        )
        for attack_key in self.attacks:
            attack = self.attacks[attack_key]
            attack_results = self._format_summary(
                cast(Union[Dict, List], batch_summarizers[attack.name].summary)
            )
            results[attack.name] = attack_results

            if len(attack_results) == 1:
                key = next(iter(attack_results))
                if attack.name not in self.summary_results:
                    self.summary_results[attack.name] = Summarizer(
                        [stat() for stat in self.aggregate_stats]
                    )
                self.summary_results[attack.name].update(
                    self.metric_aggregator(attack_results[key])
                )
            else:
                for key in attack_results:
                    summary_key = f"{attack.name} {key.title()} Attempt"
                    if summary_key not in self.summary_results:
                        self.summary_results[summary_key] = Summarizer(
                            [stat() for stat in self.aggregate_stats]
                        )
                    self.summary_results[summary_key].update(
                        self.metric_aggregator(attack_results[key])
                    )
        return results

    def summary(self) -> Dict[str, Dict[str, MetricResultType]]:
        r"""
        Returns average results over all previous batches evaluated.

        Returns:

            - **summary** Dict: str -> Dict[str, Union[Tensor, Tuple[Tensor, ...]]]:
                Dictionary containing summarized average attack results.
                Maps attack name (with "Mean Attempt", "Max Attempt" and "Min Attempt"
                suffixes if num_attempts > 1) to dictionary containing a key of "mean"
                maintaining summarized results,
                which is the running mean of results over all batches
                since construction or previous reset call. Tensor metrics are averaged
                over dimension 0 for each batch, in order to aggregte metrics collected
                per batch.
        """
        return {
            key: self._format_summary(
                cast(Union[Dict, List], self.summary_results[key].summary)
            )
            for key in self.summary_results
        }

    def reset(self) -> None:
        r"""
        Reset stored average summary results for previous batches
        """
        self.summary_results = {}



#!/usr/bin/env python3
from typing import Any, Callable, List, Tuple, Union

import torch
import torch.nn.functional as F
from captum._utils.common import (
    _format_additional_forward_args,
    _format_output,
    _format_tensor_into_tuples,
)
from captum._utils.gradient import compute_layer_gradients_and_eval
from captum._utils.typing import TargetType
from captum.attr._utils.attribution import GradientAttribution, LayerAttribution
from captum.log import log_usage
from torch import Tensor
from torch.nn import Module


class LayerGradCam(LayerAttribution, GradientAttribution):
    r"""
    Computes GradCAM attribution for chosen layer. GradCAM is designed for
    convolutional neural networks, and is usually applied to the last
    convolutional layer.

    GradCAM computes the gradients of the target output with respect to
    the given layer, averages for each output channel (dimension 2 of
    output), and multiplies the average gradient for each channel by the
    layer activations. The results are summed over all channels.

    Note that in the original GradCAM algorithm described in the paper,
    ReLU is applied to the output, returning only non-negative attributions.
    For providing more flexibility to the user, we choose to not perform the
    ReLU internally by default and return the sign information. To match the
    original GradCAM algorithm, it is necessary to pass the parameter
    relu_attributions=True to apply ReLU on the final
    attributions or alternatively only visualize the positive attributions.

    Note: this procedure sums over the second dimension (# of channels),
    so the output of GradCAM attributions will have a second
    dimension of 1, but all other dimensions will match that of the layer
    output.

    GradCAM attributions are generally upsampled and can be viewed as a
    mask to the input, since a convolutional layer output generally
    matches the input image spatially. This upsampling can be performed
    using LayerAttribution.interpolate, as shown in the example below.

    More details regarding the GradCAM method can be found in the
    original paper here:
    https://arxiv.org/abs/1610.02391
    """

    def __init__(
        self,
        forward_func: Callable,
        layer: Module,
        device_ids: Union[None, List[int]] = None,
    ) -> None:
        r"""
        Args:

            forward_func (Callable): The forward function of the model or any
                          modification of it
            layer (torch.nn.Module): Layer for which attributions are computed.
                          Output size of attribute matches this layer's output
                          dimensions, except for dimension 2, which will be 1,
                          since GradCAM sums over channels.
            device_ids (list[int]): Device ID list, necessary only if forward_func
                          applies a DataParallel model. This allows reconstruction of
                          intermediate outputs from batched results across devices.
                          If forward_func is given as the DataParallel model itself,
                          then it is not necessary to provide this argument.
        """
        LayerAttribution.__init__(self, forward_func, layer, device_ids)
        GradientAttribution.__init__(self, forward_func)

    @log_usage()
    def attribute(
        self,
        inputs: Union[Tensor, Tuple[Tensor, ...]],
        target: TargetType = None,
        additional_forward_args: Any = None,
        attribute_to_layer_input: bool = False,
        relu_attributions: bool = False,
        attr_dim_summation: bool = True,
    ) -> Union[Tensor, Tuple[Tensor, ...]]:
        r"""
        Args:

            inputs (Tensor or tuple[Tensor, ...]): Input for which attributions
                        are computed. If forward_func takes a single
                        tensor as input, a single input tensor should be provided.
                        If forward_func takes multiple tensors as input, a tuple
                        of the input tensors should be provided. It is assumed
                        that for all given input tensors, dimension 0 corresponds
                        to the number of examples, and if multiple input tensors
                        are provided, the examples must be aligned appropriately.
            target (int, tuple, Tensor, or list, optional): Output indices for
                        which gradients are computed (for classification cases,
                        this is usually the target class).
                        If the network returns a scalar value per example,
                        no target index is necessary.
                        For general 2D outputs, targets can be either:

                        - a single integer or a tensor containing a single
                          integer, which is applied to all input examples

                        - a list of integers or a 1D tensor, with length matching
                          the number of examples in inputs (dim 0). Each integer
                          is applied as the target for the corresponding example.

                        For outputs with > 2 dimensions, targets can be either:

                        - A single tuple, which contains #output_dims - 1
                          elements. This target index is applied to all examples.

                        - A list of tuples with length equal to the number of
                          examples in inputs (dim 0), and each tuple containing
                          #output_dims - 1 elements. Each tuple is applied as the
                          target for the corresponding example.

                        Default: None
            additional_forward_args (Any, optional): If the forward function
                        requires additional arguments other than the inputs for
                        which attributions should not be computed, this argument
                        can be provided. It must be either a single additional
                        argument of a Tensor or arbitrary (non-tuple) type or a
                        tuple containing multiple additional arguments including
                        tensors or any arbitrary python types. These arguments
                        are provided to forward_func in order following the
                        arguments in inputs.
                        Note that attributions are not computed with respect
                        to these arguments.
                        Default: None
            attribute_to_layer_input (bool, optional): Indicates whether to
                        compute the attributions with respect to the layer input
                        or output. If `attribute_to_layer_input` is set to True
                        then the attributions will be computed with respect to the
                        layer input, otherwise it will be computed with respect
                        to layer output.
                        Note that currently it is assumed that either the input
                        or the outputs of internal layers, depending on whether we
                        attribute to the input or output, are single tensors.
                        Support for multiple tensors will be added later.
                        Default: False
            relu_attributions (bool, optional): Indicates whether to
                        apply a ReLU operation on the final attribution,
                        returning only non-negative attributions. Setting this
                        flag to True matches the original GradCAM algorithm,
                        otherwise, by default, both positive and negative
                        attributions are returned.
                        Default: False
            attr_dim_summation (bool, optional): Indicates whether to
                        sum attributions along dimension 1 (usually channel).
                        The default (True) means to sum along dimension 1.
                        Default: True

        Returns:
            *Tensor* or *tuple[Tensor, ...]* of **attributions**:
            - **attributions** (*Tensor* or *tuple[Tensor, ...]*):
                        Attributions based on GradCAM method.
                        Attributions will be the same size as the
                        output of the given layer, except for dimension 2,
                        which will be 1 due to summing over channels.
                        Attributions are returned in a tuple if
                        the layer inputs / outputs contain multiple tensors,
                        otherwise a single tensor is returned.
        Examples::

            >>> # ImageClassifier takes a single input tensor of images Nx3x32x32,
            >>> # and returns an Nx10 tensor of class probabilities.
            >>> # It contains a layer conv4, which is an instance of nn.conv2d,
            >>> # and the output of this layer has dimensions Nx50x8x8.
            >>> # It is the last convolution layer, which is the recommended
            >>> # use case for GradCAM.
            >>> net = ImageClassifier()
            >>> layer_gc = LayerGradCam(net, net.conv4)
            >>> input = torch.randn(2, 3, 32, 32, requires_grad=True)
            >>> # Computes layer GradCAM for class 3.
            >>> # attribution size matches layer output except for dimension
            >>> # 1, so dimensions of attr would be Nx1x8x8.
            >>> attr = layer_gc.attribute(input, 3)
            >>> # GradCAM attributions are often upsampled and viewed as a
            >>> # mask to the input, since the convolutional layer output
            >>> # spatially matches the original input image.
            >>> # This can be done with LayerAttribution's interpolate method.
            >>> upsampled_attr = LayerAttribution.interpolate(attr, (32, 32))
        """
        inputs = _format_tensor_into_tuples(inputs)
        additional_forward_args = _format_additional_forward_args(
            additional_forward_args
        )
        # Returns gradient of output with respect to
        # hidden layer and hidden layer evaluated at each input.
        layer_gradients, layer_evals = compute_layer_gradients_and_eval(
            self.forward_func,
            self.layer,
            inputs,
            target,
            additional_forward_args,
            device_ids=self.device_ids,
            attribute_to_layer_input=attribute_to_layer_input,
        )

        summed_grads = tuple(
            torch.mean(
                layer_grad,
                dim=tuple(x for x in range(2, len(layer_grad.shape))),
                keepdim=True,
            )
            if len(layer_grad.shape) > 2
            else layer_grad
            for layer_grad in layer_gradients
        )

        if attr_dim_summation:
            scaled_acts = tuple(
                torch.sum(summed_grad * layer_eval, dim=1, keepdim=True)
                for summed_grad, layer_eval in zip(summed_grads, layer_evals)
            )
        else:
            scaled_acts = tuple(
                summed_grad * layer_eval
                for summed_grad, layer_eval in zip(summed_grads, layer_evals)
            )

        if relu_attributions:
            scaled_acts = tuple(F.relu(scaled_act) for scaled_act in scaled_acts)
        return _format_output(len(scaled_acts) > 1, scaled_acts)



#!/usr/bin/env fbpython
import math
from typing import cast
from unittest.mock import Mock, patch

import torch

from captum.attr._core.dataloader_attr import DataLoaderAttribution, InputRole
from captum.attr._core.feature_ablation import FeatureAblation
from parameterized import parameterized
from tests.helpers.basic import (
    assertAttributionComparision,
    assertTensorAlmostEqual,
    BaseTest,
)
from torch import Tensor
from torch.utils.data import DataLoader, TensorDataset


def sum_forward(*inps):
    inps = [torch.flatten(inp, start_dim=1) for inp in inps]
    return torch.cat(inps, dim=1).sum(1)


class Linear(torch.nn.Module):
    def __init__(self, n):
        super().__init__()
        self.linear = torch.nn.Linear(n, 1)

    def forward(self, *inps):
        inps = [torch.flatten(inp, start_dim=1) for inp in inps]
        return self.linear(torch.cat(inps, dim=1))


mock_dataset = TensorDataset(
    # iD feature
    torch.tensor(
        [
            [0.0, 0.1],
            [0.3, 0.4],
            [0.6, 0.7],
            [0.9, 1.0],
            [1.2, 1.3],
        ]
    ),
    # 2D feature
    torch.tensor(
        [
            [[0.1, 0.2], [0.3, 0.2]],
            [[0.4, 0.5], [0.3, 0.2]],
            [[0.8, 0.1], [0.2, 0.5]],
            [[1.1, 0.7], [0.1, 0.7]],
            [[0.6, 1.4], [1.2, 0.4]],
        ]
    ),
    # scalar feature or label
    torch.tensor(
        [
            [0],
            [1],
            [0],
            [0],
            [1],
        ]
    ),
)


class Test(BaseTest):
    @parameterized.expand(
        [
            (sum_forward,),
            (Linear(7),),
        ]
    )
    def test_dl_attr(self, forward) -> None:
        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        dataloader = DataLoader(mock_dataset, batch_size=2)

        dl_attributions = dl_fa.attribute(dataloader)

        # default reduce of DataLoaderAttribution works the same as concat all batches
        attr_list = []
        for batch in dataloader:
            batch_attr = fa.attribute(tuple(batch))
            attr_list.append(batch_attr)

        expected_attr = tuple(
            torch.cat(feature_attrs, dim=0) for feature_attrs in zip(*attr_list)
        )

        assertAttributionComparision(self, dl_attributions, expected_attr)

    @parameterized.expand(
        [
            (sum_forward,),
            (Linear(7),),
        ]
    )
    def test_dl_attr_with_mask(self, forward) -> None:
        # FeatureAblation does not support grouping across tensors for now
        # add such test cases after support grouping across tensors in FeatureAblation
        masks = (
            torch.tensor([[0, 0]]),
            torch.tensor([[[1, 2], [3, 2]]]),
            torch.tensor([[4]]),
        )

        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        dataloader = DataLoader(mock_dataset, batch_size=2)

        dl_attributions = dl_fa.attribute(dataloader, feature_mask=masks)

        # default reduce of DataLoaderAttribution works the same as concat all batches
        attr_list = []
        for batch in dataloader:
            batch_attr = fa.attribute(tuple(batch), feature_mask=masks)
            attr_list.append(batch_attr)

        expected_attr = tuple(
            torch.cat(feature_attrs, dim=0) for feature_attrs in zip(*attr_list)
        )

        assertAttributionComparision(self, dl_attributions, expected_attr)

    @parameterized.expand(
        [
            (sum_forward,),
            (Linear(7),),
        ]
    )
    def test_dl_attr_with_baseline(self, forward) -> None:
        baselines = (
            torch.tensor([[0, -1]]),
            1,
            0.1,
        )

        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        dataloader = DataLoader(mock_dataset, batch_size=2)

        dl_attributions = dl_fa.attribute(dataloader, baselines=baselines)

        # default reduce of DataLoaderAttribution works the same as concat all batches
        attr_list = []
        for batch in dataloader:
            batch_attr = fa.attribute(tuple(batch), baselines=baselines)
            attr_list.append(batch_attr)

        expected_attr = tuple(
            torch.cat(feature_attrs, dim=0) for feature_attrs in zip(*attr_list)
        )

        assertAttributionComparision(self, dl_attributions, expected_attr)

    def test_dl_attr_with_reduce_and_to_metric(self) -> None:
        forward = sum_forward
        func_call_counts = {
            "reduce": 0,
            "to_metric": 0,
        }

        def reduce(accum, cur_output, cur_inputs):
            func_call_counts["reduce"] += 1

            accum = {"sum": 0, "count": 0} if accum is None else accum

            accum["sum"] += cur_output.sum()
            accum["count"] += len(cur_output)

            return accum

        def to_metric(accum):
            func_call_counts["to_metric"] += 1

            self.assertEqual(isinstance(accum, dict), True)
            return torch.tensor(
                [
                    accum["sum"] / accum["count"],
                    accum["sum"],
                ]
            )

        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        batch_size = 2
        dataloader = DataLoader(mock_dataset, batch_size=batch_size)

        dl_attribution = dl_fa.attribute(
            dataloader,
            reduce=reduce,
            to_metric=to_metric,
            return_input_shape=False,
        )

        n_iters = len(dataloader)

        n_features = 7
        # after support other attr methods, this can be diff from n_features
        n_perturbations = 7
        n_passes = n_perturbations + 1  # +1 for base forward without perturbation
        n_outputs = 2  # [mean, sum]

        self.assertEqual(func_call_counts["reduce"], n_iters * n_passes)
        self.assertEqual(func_call_counts["to_metric"], n_passes)

        expected_attr_shape = (n_outputs, n_features)

        self.assertEqual(type(dl_attribution), Tensor)
        dl_attribution = cast(Tensor, dl_attribution)
        self.assertEqual(dl_attribution.shape, expected_attr_shape)

    @parameterized.expand(
        [
            ([0, 0, 0],),
            ([0, 1, 0],),
            ([0, 1, 1],),
            ([0, 1, 2],),
            ([0, 2, 2],),
        ]
    )
    def test_dl_attr_with_input_roles(self, input_roles) -> None:
        n_inputs = len(input_roles)
        n_forward_inputs = sum(1 for r in input_roles if r != InputRole.no_forward)
        n_attr_inputs = sum(1 for r in input_roles if r == InputRole.need_attr)

        def reduce(accum, cur_output, cur_inputs):
            # all inputs from dataloader should be given to reduce
            self.assertEqual(len(cur_inputs), n_inputs)

            return cur_output if accum is None else torch.cat([accum, cur_output])

        def forward(*forward_inputs):
            # inputs of InputRole.no_forward should not be passed to forward
            self.assertEqual(len(forward_inputs), n_forward_inputs)
            return sum_forward(*forward_inputs)

        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        batch_size = 2
        dataloader = DataLoader(mock_dataset, batch_size=batch_size)

        dl_attributions = dl_fa.attribute(
            dataloader,
            input_roles=input_roles,
            reduce=reduce,
        )

        # only inputs needs
        self.assertEqual(len(dl_attributions), n_attr_inputs)

        # default reduce of DataLoaderAttribution works the same as concat all batches
        attr_list = []
        for batch in dataloader:
            attr_inputs = tuple(
                _ for _, role in zip(batch, input_roles) if role == InputRole.need_attr
            )
            additional_forward_args = tuple(
                _
                for _, role in zip(batch, input_roles)
                if role == InputRole.need_forward
            )

            batch_attr = fa.attribute(
                attr_inputs, additional_forward_args=additional_forward_args
            )
            attr_list.append(batch_attr)

        expected_attr = tuple(
            torch.cat(feature_attrs, dim=0) for feature_attrs in zip(*attr_list)
        )

        assertAttributionComparision(self, dl_attributions, expected_attr)

    def test_dl_attr_not_return_input_shape(self) -> None:
        forward = sum_forward
        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        dataloader = DataLoader(mock_dataset, batch_size=2)

        dl_attribution = dl_fa.attribute(dataloader, return_input_shape=False)

        expected_attr_shape = (len(mock_dataset), 7)

        self.assertEqual(type(dl_attribution), Tensor)
        dl_attribution = cast(Tensor, dl_attribution)
        self.assertEqual(dl_attribution.shape, expected_attr_shape)

        # default reduce of DataLoaderAttribution works the same as concat all batches
        attr_list = []
        for batch in dataloader:
            batch_attr = fa.attribute(tuple(batch))
            attr_list.append(batch_attr)

        expected_attr = torch.cat(
            [
                # flatten feature dim
                torch.cat(feature_attrs, dim=0).flatten(start_dim=1)
                for feature_attrs in zip(*attr_list)
            ],
            dim=1,
        )

        assertTensorAlmostEqual(self, dl_attribution, expected_attr)

    def test_dl_attr_with_mask_not_return_input_shape(self) -> None:
        forward = sum_forward
        masks = (
            torch.tensor([[0, 0]]),
            torch.tensor([[[1, 2], [3, 2]]]),
            torch.tensor([[4]]),
        )

        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        dataloader = DataLoader(mock_dataset, batch_size=2)

        dl_attribution = dl_fa.attribute(
            dataloader, feature_mask=masks, return_input_shape=False
        )

        expected_attr_shape = (len(mock_dataset), 5)

        self.assertEqual(type(dl_attribution), Tensor)
        dl_attribution = cast(Tensor, dl_attribution)
        self.assertEqual(dl_attribution.shape, expected_attr_shape)

    @parameterized.expand([(2,), (3,), (4,)])
    def test_dl_attr_with_perturb_per_pass(self, perturb_per_pass) -> None:
        forward = sum_forward

        fa = FeatureAblation(forward)
        dl_fa = DataLoaderAttribution(fa)

        mock_dl_iter = Mock(wraps=DataLoader.__iter__)

        with patch.object(DataLoader, "__iter__", lambda self: mock_dl_iter(self)):
            dataloader = DataLoader(mock_dataset, batch_size=2)

            dl_attributions = dl_fa.attribute(
                dataloader, perturbations_per_pass=perturb_per_pass
            )

        n_features = 7
        # 2 extra iter calls: get one input for format; get unperturbed output
        n_iter_overhead = 2

        self.assertEqual(
            mock_dl_iter.call_count,
            math.ceil(n_features / perturb_per_pass) + n_iter_overhead,
        )

        # default reduce of DataLoaderAttribution works the same as concat all batches
        attr_list = []
        for batch in dataloader:
            batch_attr = fa.attribute(tuple(batch))
            attr_list.append(batch_attr)

        expected_attr = tuple(
            torch.cat(feature_attrs, dim=0) for feature_attrs in zip(*attr_list)
        )

        assertAttributionComparision(self, dl_attributions, expected_attr)



#!/usr/bin/env python3
from typing import List, Tuple

import torch
from captum.attr._core.feature_permutation import _permute_feature, FeaturePermutation
from tests.helpers.basic import assertTensorAlmostEqual, BaseTest
from tests.helpers.basic_models import BasicModelWithSparseInputs
from torch import Tensor


class Test(BaseTest):
    def _check_features_are_permuted(
        self, inp: Tensor, perm_inp: Tensor, mask: Tensor
    ) -> None:
        permuted_features = mask.expand_as(inp[0])
        unpermuted_features = permuted_features.bitwise_not()

        self.assertTrue(inp.dtype == perm_inp.dtype)
        self.assertTrue(inp.shape == perm_inp.shape)
        self.assertTrue(
            (inp[:, permuted_features] != perm_inp[:, permuted_features]).any()
        )
        self.assertTrue(
            (inp[:, unpermuted_features] == perm_inp[:, unpermuted_features]).all()
        )

    def _check_perm_fn_with_mask(self, inp: Tensor, mask: Tensor) -> None:
        perm_inp = _permute_feature(inp, mask)
        self._check_features_are_permuted(inp, perm_inp, mask)

    def test_perm_fn_single_feature(self) -> None:
        batch_size = 2
        sizes_to_test: List[Tuple[int, ...]] = [(10,), (4, 5), (3, 4, 5)]
        for inp_size in sizes_to_test:
            inp = torch.randn((batch_size,) + inp_size)
            flat_mask = torch.zeros_like(inp[0]).flatten().bool()

            num_features = inp.numel() // batch_size
            for i in range(num_features):
                flat_mask[i] = 1
                self._check_perm_fn_with_mask(inp, flat_mask.view_as(inp[0]))
                flat_mask[i] = 0

    def test_perm_fn_broadcastable_masks(self) -> None:
        batch_size = 5
        inp_size = (3, 20, 30)

        inp = torch.randn((batch_size,) + inp_size)

        # To be broadcastable dimensions have
        # match from end to beginning, by equalling 1 or the dim.
        #
        # If a dimension is missing then it must be the
        # last dim provided (from right to left). The missing
        # dimensions are implied to be = 1
        #
        # Here I write them explicitly for clarity
        mask_sizes: List[Tuple[int, ...]] = [
            # dims = 1
            (1, 20, 30),
            (3, 1, 30),
            (3, 20, 1),
            (1, 1, 30),
            (1, 20, 1),
            # missing
            (1,),  # empty set (all features)
            (30,),
            (20, 30),
            (3, 20, 30),
        ]

        for mask_size in mask_sizes:
            mask = torch.randint(0, 2, mask_size).bool()
            self.assertTrue(mask.shape == mask_size)

            self._check_perm_fn_with_mask(inp, mask)

    def test_single_input(self) -> None:
        batch_size = 2
        input_size = (6,)
        constant_value = 10000

        def forward_func(x: Tensor) -> Tensor:
            return x.sum(dim=-1)

        feature_importance = FeaturePermutation(forward_func=forward_func)

        inp = torch.randn((batch_size,) + input_size)

        inp[:, 0] = constant_value
        zeros = torch.zeros_like(inp[:, 0])

        attribs = feature_importance.attribute(inp)

        self.assertTrue(attribs.squeeze(0).size() == (batch_size,) + input_size)
        assertTensorAlmostEqual(self, attribs[:, 0], zeros, delta=0.05, mode="max")
        self.assertTrue((attribs[:, 1 : input_size[0]].abs() > 0).all())

    def test_multi_input(self) -> None:
        batch_size = 20
        inp1_size = (5, 2)
        inp2_size = (5, 3)

        labels = torch.randn(batch_size)

        def forward_func(*x: Tensor) -> Tensor:
            y = torch.zeros(x[0].shape[0:2])
            for xx in x:
                y += xx[:, :, 0] * xx[:, :, 1]
            y = y.sum(dim=-1)

            return torch.mean((y - labels) ** 2)

        feature_importance = FeaturePermutation(forward_func=forward_func)

        inp = (
            torch.randn((batch_size,) + inp1_size),
            torch.randn((batch_size,) + inp2_size),
        )

        feature_mask = (
            torch.arange(inp[0][0].numel()).view_as(inp[0][0]).unsqueeze(0),
            torch.arange(inp[1][0].numel()).view_as(inp[1][0]).unsqueeze(0),
        )

        inp[1][:, :, 1] = 4
        attribs = feature_importance.attribute(inp, feature_mask=feature_mask)

        self.assertTrue(isinstance(attribs, tuple))
        self.assertTrue(len(attribs) == 2)

        self.assertTrue(attribs[0].squeeze(0).size() == inp1_size)
        self.assertTrue(attribs[1].squeeze(0).size() == inp2_size)

        self.assertTrue((attribs[1][:, :, 1] == 0).all())
        self.assertTrue((attribs[1][:, :, 2] == 0).all())

        self.assertTrue((attribs[0] != 0).all())
        self.assertTrue((attribs[1][:, :, 0] != 0).all())

    def test_mulitple_perturbations_per_eval(self) -> None:
        perturbations_per_eval = 4
        batch_size = 2
        input_size = (4,)

        inp = torch.randn((batch_size,) + input_size)

        def forward_func(x):
            return 1 - x

        target = 1
        feature_importance = FeaturePermutation(forward_func=forward_func)

        attribs = feature_importance.attribute(
            inp, perturbations_per_eval=perturbations_per_eval, target=target
        )
        self.assertTrue(attribs.size() == (batch_size,) + input_size)

        for i in range(inp.size(1)):
            if i == target:
                continue
            assertTensorAlmostEqual(
                self, attribs[:, i], torch.zeros_like(attribs[:, i])
            )

        y = forward_func(inp)
        actual_diff = torch.stack([(y[0] - y[1])[target], (y[1] - y[0])[target]])
        assertTensorAlmostEqual(self, attribs[:, target], actual_diff)

    def test_broadcastable_masks(self) -> None:
        # integration test to ensure that
        # permutation function works with custom masks
        def forward_func(x: Tensor) -> Tensor:
            return x.view(x.shape[0], -1).sum(dim=-1)

        batch_size = 2
        inp = torch.randn((batch_size,) + (3, 4, 4))

        feature_importance = FeaturePermutation(forward_func=forward_func)

        masks = [
            torch.tensor([0]),
            torch.tensor([[0, 1, 2, 3]]),
            torch.tensor([[[0, 1, 2, 3], [3, 3, 4, 5], [6, 6, 4, 6], [7, 8, 9, 10]]]),
        ]

        for mask in masks:
            attribs = feature_importance.attribute(inp, feature_mask=mask)

            self.assertTrue(attribs is not None)
            self.assertTrue(attribs.shape == inp.shape)

            fm = mask.expand_as(inp[0])

            features = set(mask.flatten())
            for feature in features:
                m = (fm == feature).bool()
                attribs_for_feature = attribs[:, m]
                assertTensorAlmostEqual(
                    self,
                    attribs_for_feature[0],
                    -attribs_for_feature[1],
                    delta=0.05,
                    mode="max",
                )

    def test_empty_sparse_features(self) -> None:
        model = BasicModelWithSparseInputs()
        inp1 = torch.tensor([[1.0, -2.0, 3.0], [2.0, -1.0, 3.0]])
        inp2 = torch.tensor([])

        # test empty sparse tensor
        feature_importance = FeaturePermutation(model)
        attr1, attr2 = feature_importance.attribute((inp1, inp2))
        self.assertEqual(attr1.shape, (1, 3))
        self.assertEqual(attr2.shape, (1,))

    def test_sparse_features(self) -> None:
        model = BasicModelWithSparseInputs()
        inp1 = torch.tensor([[1.0, -2.0, 3.0], [2.0, -1.0, 3.0]])
        # Length of sparse index list may not match # of examples
        inp2 = torch.tensor([1, 7, 2, 4, 5, 3, 6])

        feature_importance = FeaturePermutation(model)
        total_attr1, total_attr2 = feature_importance.attribute((inp1, inp2))

        for _ in range(50):
            attr1, attr2 = feature_importance.attribute((inp1, inp2))
            total_attr1 += attr1
            total_attr2 += attr2
        total_attr1 /= 50
        total_attr2 /= 50
        self.assertEqual(total_attr2.shape, (1,))
        assertTensorAlmostEqual(self, total_attr1, torch.zeros_like(total_attr1))
        assertTensorAlmostEqual(self, total_attr2, [-6.0], delta=0.2)



import os
import tempfile
from typing import Callable, cast, Optional

import torch
import torch.nn as nn
from captum.influence._core.tracincp import TracInCP
from captum.influence._core.tracincp_fast_rand_proj import (
    TracInCPFast,
    TracInCPFastRandProj,
)
from parameterized import parameterized
from tests.helpers.basic import assertTensorAlmostEqual, BaseTest
from tests.influence._utils.common import (
    _isSorted,
    _wrap_model_in_dataparallel,
    build_test_name_func,
    CoefficientNet,
    DataInfluenceConstructor,
    IdentityDataset,
    RangeDataset,
)


class TestTracInRegression(BaseTest):
    def _test_tracin_regression_setup(
        self, tmpdir: str, features: int, use_gpu: bool = False
    ):
        low = 1
        high = 17
        dataset = RangeDataset(low, high, features, use_gpu)
        net = CoefficientNet(in_features=features)

        checkpoint_name = "-".join(["checkpoint-reg", "0" + ".pt"])
        torch.save(net.state_dict(), os.path.join(tmpdir, checkpoint_name))

        weights = [0.4379, 0.1653, 0.5132, 0.3651, 0.9992]

        for i, weight in enumerate(weights):
            net.fc1.weight.data.fill_(weight)
            net_adjusted = _wrap_model_in_dataparallel(net) if use_gpu else net
            checkpoint_name = "-".join(["checkpoint-reg", str(i + 1) + ".pt"])
            torch.save(net_adjusted.state_dict(), os.path.join(tmpdir, checkpoint_name))

        return dataset, net_adjusted

    use_gpu_list = (
        [True, False]
        if torch.cuda.is_available() and torch.cuda.device_count() != 0
        else [False]
    )

    param_list = []
    for use_gpu in use_gpu_list:
        for dim in [1, 20]:
            for (mode, reduction, constructor) in [
                (
                    "check_idx",
                    "none",
                    DataInfluenceConstructor(TracInCP, name="TracInCP_all_layers"),
                ),
                (
                    "check_idx",
                    "none",
                    DataInfluenceConstructor(
                        TracInCP,
                        name="TracInCP_fc1",
                        layers=["module.fc1"] if use_gpu else ["fc1"],
                    ),
                ),
                (
                    "sample_wise_trick",
                    None,
                    DataInfluenceConstructor(TracInCP, name="TracInCP_fc1"),
                ),
                (
                    "check_idx",
                    "sum",
                    DataInfluenceConstructor(
                        TracInCPFast, name="TracInCPFast_last_fc_layer"
                    ),
                ),
                (
                    "check_idx",
                    "sum",
                    DataInfluenceConstructor(
                        TracInCPFastRandProj, name="TracInCPFast_last_fc_layer"
                    ),
                ),
                (
                    "check_idx",
                    "mean",
                    DataInfluenceConstructor(
                        TracInCPFast, name="TracInCPFast_last_fc_layer"
                    ),
                ),
                (
                    "check_idx",
                    "mean",
                    DataInfluenceConstructor(
                        TracInCPFastRandProj, name="TracInCPFastRandProj_last_fc_layer"
                    ),
                ),
                (
                    "check_idx",
                    "sum",
                    DataInfluenceConstructor(
                        TracInCPFastRandProj,
                        name="TracInCPFastRandProj1DimensionalProjection_last_fc_layer",
                        projection_dim=1,
                    ),
                ),
                (
                    "check_idx",
                    "mean",
                    DataInfluenceConstructor(
                        TracInCPFast,
                        name="TracInCPFastDuplicateLossFn",
                        duplicate_loss_fn=True,
                    ),
                ),  # add a test where `duplicate_loss_fn` is True
                (
                    "check_idx",
                    "mean",
                    DataInfluenceConstructor(
                        TracInCPFastRandProj,
                        name="TracInCPFastRandProjDuplicateLossFn",
                        duplicate_loss_fn=True,
                    ),  # add a test where `duplicate_loss_fn` is True
                ),
            ]:
                if not (mode == "sample_wise_trick" and use_gpu):
                    param_list.append((reduction, constructor, mode, dim, use_gpu))

    @parameterized.expand(
        param_list,
        name_func=build_test_name_func(args_to_skip=["reduction"]),
    )
    def test_tracin_regression(
        self,
        reduction: Optional[str],
        tracin_constructor: Callable,
        mode: str,
        features: int,
        use_gpu: bool,
    ) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:

            batch_size = 4

            dataset, net = self._test_tracin_regression_setup(
                tmpdir,
                features,
                use_gpu,
            )  # and not mode == 'sample_wise_trick'

            # check influence scores of training data

            train_inputs = dataset.samples
            train_labels = dataset.labels

            test_inputs = (
                torch.arange(17, 33, dtype=torch.float).unsqueeze(1).repeat(1, features)
            )

            if use_gpu:
                test_inputs = test_inputs.cuda()

            test_labels = test_inputs

            self.assertTrue(callable(tracin_constructor))

            if mode == "check_idx":

                self.assertTrue(isinstance(reduction, str))
                criterion = nn.MSELoss(reduction=cast(str, reduction))

                tracin = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                )

                train_scores = tracin.influence((train_inputs, train_labels))
                idx, _ = tracin.influence(
                    (train_inputs, train_labels), k=len(dataset), proponents=True
                )
                # check that top influence is one with maximal value
                # (and hence gradient)
                for i in range(len(idx)):
                    self.assertEqual(idx[i][0], 15)

                # check influence scores of test data
                test_scores = tracin.influence((test_inputs, test_labels))
                idx, _ = tracin.influence(
                    (test_inputs, test_labels), k=len(test_inputs), proponents=True
                )
                # check that top influence is one with maximal value
                # (and hence gradient)
                for i in range(len(idx)):
                    self.assertTrue(_isSorted(idx[i]))

            if mode == "sample_wise_trick":

                criterion = nn.MSELoss(reduction="none")

                tracin = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                    sample_wise_grads_per_batch=False,
                )

                # With sample-wise trick
                criterion = nn.MSELoss(reduction="sum")
                tracin_sample_wise_trick = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                    sample_wise_grads_per_batch=True,
                )

                train_scores = tracin.influence((train_inputs, train_labels))
                train_scores_sample_wise_trick = tracin_sample_wise_trick.influence(
                    (train_inputs, train_labels)
                )
                assertTensorAlmostEqual(
                    self, train_scores, train_scores_sample_wise_trick
                )

                test_scores = tracin.influence((test_inputs, test_labels))
                test_scores_sample_wise_trick = tracin_sample_wise_trick.influence(
                    (test_inputs, test_labels)
                )
                assertTensorAlmostEqual(
                    self, test_scores, test_scores_sample_wise_trick
                )

    @parameterized.expand(
        [
            (
                "sum",
                DataInfluenceConstructor(TracInCP, sample_wise_grads_per_batch=True),
            ),
            ("sum", DataInfluenceConstructor(TracInCPFast)),
            ("sum", DataInfluenceConstructor(TracInCPFastRandProj)),
            ("mean", DataInfluenceConstructor(TracInCPFast)),
            ("mean", DataInfluenceConstructor(TracInCPFastRandProj)),
        ],
        name_func=build_test_name_func(),
    )
    def test_tracin_regression_1D_numerical(
        self, reduction: str, tracin_constructor: Callable
    ) -> None:

        low = 1
        high = 17
        features = 1
        dataset = RangeDataset(low, high, features)
        net = CoefficientNet()
        self.assertTrue(isinstance(reduction, str))
        criterion = nn.MSELoss(reduction=cast(str, reduction))
        batch_size = 4
        weights = [0.4379, 0.1653, 0.5132, 0.3651, 0.9992]

        train_inputs = dataset.samples
        train_labels = dataset.labels

        with tempfile.TemporaryDirectory() as tmpdir:

            for i, weight in enumerate(weights):
                net.fc1.weight.data.fill_(weight)
                checkpoint_name = "-".join(["checkpoint-reg", str(i + 1) + ".pt"])
                torch.save(net.state_dict(), os.path.join(tmpdir, checkpoint_name))

            self.assertTrue(callable(tracin_constructor))
            tracin = tracin_constructor(
                net,
                dataset,
                tmpdir,
                batch_size,
                criterion,
            )

            train_scores = tracin.influence((train_inputs, train_labels), k=None)

            r"""
            Derivation for gradient / resulting TracIn score:

            For each checkpoint: $y = Wx,$ and $loss = (y - label)^2.$ Recall for this
            test case, there is no activation on y. For this example, $label = x.$

            Fast Rand Proj gives $\nabla_W loss = \nabla_y loss (x^T).$ We have $x$ and
            y as scalars so we can simply multiply. So then,

            \[\nabla_y loss * x = 2(y-x)*x = 2(Wx -x)*x = 2x^2 (w - 1).\]

            And we simply multiply these for x, x'. In this case, $x, x' \in [1..16]$.
            """
            for i in range(train_scores.shape[0]):
                for j in range(len(train_scores[0])):
                    _weights = torch.Tensor(weights)
                    num = 2 * (i + 1) * (i + 1) * (_weights - 1)
                    num *= 2 * (j + 1) * (j + 1) * (_weights - 1)
                    assertTensorAlmostEqual(
                        self, torch.sum(num), train_scores[i][j], delta=0.1
                    )

    def _test_tracin_identity_regression_setup(self, tmpdir: str):
        num_features = 7
        dataset = IdentityDataset(num_features)
        net = CoefficientNet()

        num_checkpoints = 5

        for i in range(num_checkpoints):
            net.fc1.weight.data = torch.rand((1, num_features))
            checkpoint_name = "-".join(["checkpoint-reg", str(i) + ".pt"])
            torch.save(net.state_dict(), os.path.join(tmpdir, checkpoint_name))

        return dataset, net

    @parameterized.expand(
        [
            ("check_idx", "none", DataInfluenceConstructor(TracInCP)),
            ("check_idx", "none", DataInfluenceConstructor(TracInCP, layers=["fc1"])),
            ("sample_wise_trick", None, DataInfluenceConstructor(TracInCP)),
            (
                "sample_wise_trick",
                None,
                DataInfluenceConstructor(TracInCP, layers=["fc1"]),
            ),
            ("check_idx", "sum", DataInfluenceConstructor(TracInCPFast)),
            ("check_idx", "sum", DataInfluenceConstructor(TracInCPFastRandProj)),
            ("check_idx", "mean", DataInfluenceConstructor(TracInCPFast)),
            ("check_idx", "mean", DataInfluenceConstructor(TracInCPFastRandProj)),
        ],
        name_func=build_test_name_func(),
    )
    def test_tracin_identity_regression(
        self, mode: str, reduction: Optional[str], tracin_constructor: Callable
    ) -> None:
        """
        This test uses a linear model with positive coefficients, where input feature
        matrix is the identity matrix.  Since the dot product between 2 different
        training instances is always 0, when calculating influence scores on the
        training data, only self influence scores will be nonzero.  Since the linear
        model has positive coefficients, self influence scores will be positive.
        Thus, the training instance with the largest influence on another training
        instance is itself.
        """

        with tempfile.TemporaryDirectory() as tmpdir:

            batch_size = 4

            dataset, net = self._test_tracin_identity_regression_setup(tmpdir)

            train_inputs = dataset.samples
            train_labels = dataset.labels

            self.assertTrue(callable(tracin_constructor))

            if mode == "check_idx":

                self.assertTrue(isinstance(reduction, str))
                criterion = nn.MSELoss(reduction=cast(str, reduction))

                tracin = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                )

                # check influence scores of training data

                train_scores = tracin.influence((train_inputs, train_labels))
                idx, _ = tracin.influence(
                    (train_inputs, train_labels), k=len(dataset), proponents=True
                )

                # check that top influence for an instance is itself
                for i in range(len(idx)):
                    self.assertEqual(idx[i][0], i)

            if mode == "sample_wise_trick":

                criterion = nn.MSELoss(reduction="none")

                tracin = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                    sample_wise_grads_per_batch=False,
                )

                # With sample-wise trick
                criterion = nn.MSELoss(reduction="sum")
                tracin_sample_wise_trick = tracin_constructor(
                    net,
                    dataset,
                    tmpdir,
                    batch_size,
                    criterion,
                    sample_wise_grads_per_batch=True,
                )

                train_scores = tracin.influence((train_inputs, train_labels))
                train_scores_tracin_sample_wise_trick = (
                    tracin_sample_wise_trick.influence((train_inputs, train_labels))
                )
                assertTensorAlmostEqual(
                    self, train_scores, train_scores_tracin_sample_wise_trick
                )

    @parameterized.expand(
        [
            ("none", "none", DataInfluenceConstructor(TracInCP)),
            (
                "mean",
                "mean",
                DataInfluenceConstructor(TracInCP, sample_wise_grads_per_batch=True),
            ),
            ("sum", "sum", DataInfluenceConstructor(TracInCPFast)),
            ("mean", "mean", DataInfluenceConstructor(TracInCPFast)),
            ("sum", "sum", DataInfluenceConstructor(TracInCPFastRandProj)),
            ("mean", "mean", DataInfluenceConstructor(TracInCPFastRandProj)),
        ],
        name_func=build_test_name_func(),
    )
    def test_tracin_constant_test_loss_fn(
        self,
        reduction: Optional[str],
        test_reduction: Optional[str],
        tracin_constructor: Callable,
    ) -> None:
        """
        All implementations of `TracInCPBase` can accept `test_loss_fn` in
        initialization, which sets the loss function applied to test examples, which
        can thus be different from the loss function applied to training examples.
        This test passes `test_loss_fn` to be a constant function. Then, the influence
        scores should all be 0, because gradients w.r.t. `test_loss_fn` will all be 0.
        It re-uses the dataset and model from `test_tracin_identity_regression`.

        The reduction for `loss_fn` and `test_loss_fn` initialization arguments is
        the same for all parameterized tests, for simplicity, and also because for
        `TracInCP`, both loss functions must both be reduction loss functions (i.e.
        reduction is "mean" or "sum"), or both be per-example loss functions (i.e.
        reduction is "none"). Recall that for `TracInCP`, the
        `sample_wise_grads_per_batch` initialization argument determines which of
        those cases holds.
        """
        with tempfile.TemporaryDirectory() as tmpdir:

            batch_size = 4

            dataset, net = self._test_tracin_identity_regression_setup(tmpdir)

            train_inputs = dataset.samples
            train_labels = dataset.labels

            self.assertTrue(callable(tracin_constructor))

            self.assertTrue(isinstance(reduction, str))
            criterion = nn.MSELoss(reduction=cast(str, reduction))

            # the output of `net`, i.e. `input` for the loss functions below, is a
            # batch_size x 1 2D tensor
            if test_reduction == "none":
                # loss function returns 1D tensor of all 0's, so is constant
                def test_loss_fn(input, target):
                    return input.squeeze() * 0.0

            elif test_reduction in ["sum", "mean"]:
                # loss function returns scalar tensor of all 0's, so is constant
                def test_loss_fn(input, target):
                    return input.mean() * 0.0

            tracin = tracin_constructor(
                net,
                dataset,
                tmpdir,
                batch_size,
                criterion,
                test_loss_fn=test_loss_fn,
            )

            # check influence scores of training data. they should all be 0
            train_scores = tracin.influence((train_inputs, train_labels), k=None)
            assertTensorAlmostEqual(self, train_scores, torch.zeros(train_scores.shape))



#!/usr/bin/env python3
from enum import Enum
from typing import Callable, List, Tuple

import torch


class Riemann(Enum):
    left = 1
    right = 2
    middle = 3
    trapezoid = 4


SUPPORTED_RIEMANN_METHODS = [
    "riemann_left",
    "riemann_right",
    "riemann_middle",
    "riemann_trapezoid",
]

SUPPORTED_METHODS = SUPPORTED_RIEMANN_METHODS + ["gausslegendre"]


def approximation_parameters(
    method: str,
) -> Tuple[Callable[[int], List[float]], Callable[[int], List[float]]]:
    r"""Retrieves parameters for the input approximation `method`

    Args:
        method (str): The name of the approximation method. Currently only `riemann`
                and gauss legendre are
    """
    if method in SUPPORTED_RIEMANN_METHODS:
        return riemann_builders(method=Riemann[method.split("_")[-1]])
    if method == "gausslegendre":
        return gauss_legendre_builders()
    raise ValueError("Invalid integral approximation method name: {}".format(method))


def riemann_builders(
    method: Riemann = Riemann.trapezoid,
) -> Tuple[Callable[[int], List[float]], Callable[[int], List[float]]]:
    r"""Step sizes are identical and alphas are scaled in [0, 1]

    Args:

         method (Riemann): `left`, `right`, `middle` and `trapezoid` riemann

    Returns:
        2-element tuple of **step_sizes**, **alphas**:
        - **step_sizes** (*Callable*):
                    `step_sizes` takes the number of steps as an
                    input argument and returns an array of steps sizes which
                    sum is smaller than or equal to one.

        - **alphas** (*Callable*):
                    `alphas` takes the number of steps as an input argument
                    and returns the multipliers/coefficients for the inputs
                    of integrand in the range of [0, 1]

    """

    def step_sizes(n: int) -> List[float]:
        assert n > 1, "The number of steps has to be larger than one"
        deltas = [1 / n] * n
        if method == Riemann.trapezoid:
            deltas[0] /= 2
            deltas[-1] /= 2
        return deltas

    def alphas(n: int) -> List[float]:
        assert n > 1, "The number of steps has to be larger than one"
        if method == Riemann.trapezoid:
            return torch.linspace(0, 1, n).tolist()
        elif method == Riemann.left:
            return torch.linspace(0, 1 - 1 / n, n).tolist()
        elif method == Riemann.middle:
            return torch.linspace(1 / (2 * n), 1 - 1 / (2 * n), n).tolist()
        elif method == Riemann.right:
            return torch.linspace(1 / n, 1, n).tolist()
        else:
            raise AssertionError("Provided Reimann approximation method is not valid.")
        # This is not a standard riemann method but in many cases it
        # leades to faster approaximation. Test cases for small number of steps
        # do not make sense but for larger number of steps the approximation is
        # better therefore leaving this option available
        # if method == 'riemann_include_endpoints':
        #     return [i / (n - 1) for i in range(n)]

    return step_sizes, alphas


def gauss_legendre_builders() -> Tuple[
    Callable[[int], List[float]], Callable[[int], List[float]]
]:
    r"""Numpy's `np.polynomial.legendre` function helps to compute step sizes
    and alpha coefficients using gauss-legendre quadrature rule.
    Since numpy returns the integration parameters in different scales we need to
    rescale them to adjust to the desired scale.

    Gauss Legendre quadrature rule for approximating the integrals was originally
    proposed by [Xue Feng and her intern Hauroun Habeeb]
    (https://research.fb.com/people/feng-xue/).

    Returns:
        2-element tuple of **step_sizes**, **alphas**:
        - **step_sizes** (*Callable*):
                    `step_sizes` takes the number of steps as an
                    input argument and returns an array of steps sizes which
                    sum is smaller than or equal to one.

        - **alphas** (*Callable*):
                    `alphas` takes the number of steps as an input argument
                    and returns the multipliers/coefficients for the inputs
                    of integrand in the range of [0, 1]

    """

    # allow using riemann even without np
    import numpy as np

    def step_sizes(n: int) -> List[float]:
        assert n > 0, "The number of steps has to be larger than zero"
        # Scaling from 2 to 1
        return list(0.5 * np.polynomial.legendre.leggauss(n)[1])

    def alphas(n: int) -> List[float]:
        assert n > 0, "The number of steps has to be larger than zero"
        # Scaling from [-1, 1] to [0, 1]
        return list(0.5 * (1 + np.polynomial.legendre.leggauss(n)[0]))

    return step_sizes, alphas



#!/usr/bin/env python3
import typing
from typing import Any, Callable, List, Tuple, Union

import torch
from captum._utils.common import (
    _expand_additional_forward_args,
    _expand_target,
    _format_additional_forward_args,
    _format_output,
    _is_tuple,
)
from captum._utils.typing import (
    BaselineType,
    Literal,
    TargetType,
    TensorOrTupleOfTensorsGeneric,
)
from captum.attr._utils.approximation_methods import approximation_parameters
from captum.attr._utils.attribution import GradientAttribution
from captum.attr._utils.batching import _batch_attribution
from captum.attr._utils.common import (
    _format_input_baseline,
    _reshape_and_sum,
    _validate_input,
)
from captum.log import log_usage
from torch import Tensor


class IntegratedGradients(GradientAttribution):
    r"""
    Integrated Gradients is an axiomatic model interpretability algorithm that
    assigns an importance score to each input feature by approximating the
    integral of gradients of the model's output with respect to the inputs
    along the path (straight line) from given baselines / references to inputs.

    Baselines can be provided as input arguments to attribute method.
    To approximate the integral we can choose to use either a variant of
    Riemann sum or Gauss-Legendre quadrature rule.

    More details regarding the integrated gradients method can be found in the
    original paper:
    https://arxiv.org/abs/1703.01365

    """

    def __init__(
        self,
        forward_func: Callable,
        multiply_by_inputs: bool = True,
    ) -> None:
        r"""
        Args:

            forward_func (Callable): The forward function of the model or any
                    modification of it
            multiply_by_inputs (bool, optional): Indicates whether to factor
                    model inputs' multiplier in the final attribution scores.
                    In the literature this is also known as local vs global
                    attribution. If inputs' multiplier isn't factored in,
                    then that type of attribution method is also called local
                    attribution. If it is, then that type of attribution
                    method is called global.
                    More detailed can be found here:
                    https://arxiv.org/abs/1711.06104

                    In case of integrated gradients, if `multiply_by_inputs`
                    is set to True, final sensitivity scores are being multiplied by
                    (inputs - baselines).
        """
        GradientAttribution.__init__(self, forward_func)
        self._multiply_by_inputs = multiply_by_inputs

    # The following overloaded method signatures correspond to the case where
    # return_convergence_delta is False, then only attributions are returned,
    # and when return_convergence_delta is True, the return type is
    # a tuple with both attributions and deltas.
    @typing.overload
    def attribute(
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: BaselineType = None,
        target: TargetType = None,
        additional_forward_args: Any = None,
        n_steps: int = 50,
        method: str = "gausslegendre",
        internal_batch_size: Union[None, int] = None,
        return_convergence_delta: Literal[False] = False,
    ) -> TensorOrTupleOfTensorsGeneric:
        ...

    @typing.overload
    def attribute(
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: BaselineType = None,
        target: TargetType = None,
        additional_forward_args: Any = None,
        n_steps: int = 50,
        method: str = "gausslegendre",
        internal_batch_size: Union[None, int] = None,
        *,
        return_convergence_delta: Literal[True],
    ) -> Tuple[TensorOrTupleOfTensorsGeneric, Tensor]:
        ...

    @log_usage()
    def attribute(  # type: ignore
        self,
        inputs: TensorOrTupleOfTensorsGeneric,
        baselines: BaselineType = None,
        target: TargetType = None,
        additional_forward_args: Any = None,
        n_steps: int = 50,
        method: str = "gausslegendre",
        internal_batch_size: Union[None, int] = None,
        return_convergence_delta: bool = False,
    ) -> Union[
        TensorOrTupleOfTensorsGeneric, Tuple[TensorOrTupleOfTensorsGeneric, Tensor]
    ]:
        r"""
        This method attributes the output of the model with given target index
        (in case it is provided, otherwise it assumes that output is a
        scalar) to the inputs of the model using the approach described above.

        In addition to that it also returns, if `return_convergence_delta` is
        set to True, integral approximation delta based on the completeness
        property of integrated gradients.

        Args:

            inputs (Tensor or tuple[Tensor, ...]): Input for which integrated
                        gradients are computed. If forward_func takes a single
                        tensor as input, a single input tensor should be provided.
                        If forward_func takes multiple tensors as input, a tuple
                        of the input tensors should be provided. It is assumed
                        that for all given input tensors, dimension 0 corresponds
                        to the number of examples, and if multiple input tensors
                        are provided, the examples must be aligned appropriately.
            baselines (scalar, Tensor, tuple of scalar, or Tensor, optional):
                        Baselines define the starting point from which integral
                        is computed and can be provided as:

                        - a single tensor, if inputs is a single tensor, with
                          exactly the same dimensions as inputs or the first
                          dimension is one and the remaining dimensions match
                          with inputs.

                        - a single scalar, if inputs is a single tensor, which will
                          be broadcasted for each input value in input tensor.

                        - a tuple of tensors or scalars, the baseline corresponding
                          to each tensor in the inputs' tuple can be:

                          - either a tensor with matching dimensions to
                            corresponding tensor in the inputs' tuple
                            or the first dimension is one and the remaining
                            dimensions match with the corresponding
                            input tensor.

                          - or a scalar, corresponding to a tensor in the
                            inputs' tuple. This scalar value is broadcasted
                            for corresponding input tensor.

                        In the cases when `baselines` is not provided, we internally
                        use zero scalar corresponding to each input tensor.

                        Default: None
            target (int, tuple, Tensor, or list, optional): Output indices for
                        which gradients are computed (for classification cases,
                        this is usually the target class).
                        If the network returns a scalar value per example,
                        no target index is necessary.
                        For general 2D outputs, targets can be either:

                        - a single integer or a tensor containing a single
                          integer, which is applied to all input examples

                        - a list of integers or a 1D tensor, with length matching
                          the number of examples in inputs (dim 0). Each integer
                          is applied as the target for the corresponding example.

                        For outputs with > 2 dimensions, targets can be either:

                        - A single tuple, which contains #output_dims - 1
                          elements. This target index is applied to all examples.

                        - A list of tuples with length equal to the number of
                          examples in inputs (dim 0), and each tuple containing
                          #output_dims - 1 elements. Each tuple is applied as the
                          target for the corresponding example.

                        Default: None
            additional_forward_args (Any, optional): If the forward function
                        requires additional arguments other than the inputs for
                        which attributions should not be computed, this argument
                        can be provided. It must be either a single additional
                        argument of a Tensor or arbitrary (non-tuple) type or a
                        tuple containing multiple additional arguments including
                        tensors or any arbitrary python types. These arguments
                        are provided to forward_func in order following the
                        arguments in inputs.
                        For a tensor, the first dimension of the tensor must
                        correspond to the number of examples. It will be
                        repeated for each of `n_steps` along the integrated
                        path. For all other types, the given argument is used
                        for all forward evaluations.
                        Note that attributions are not computed with respect
                        to these arguments.
                        Default: None
            n_steps (int, optional): The number of steps used by the approximation
                        method. Default: 50.
            method (str, optional): Method for approximating the integral,
                        one of `riemann_right`, `riemann_left`, `riemann_middle`,
                        `riemann_trapezoid` or `gausslegendre`.
                        Default: `gausslegendre` if no method is provided.
            internal_batch_size (int, optional): Divides total #steps * #examples
                        data points into chunks of size at most internal_batch_size,
                        which are computed (forward / backward passes)
                        sequentially. internal_batch_size must be at least equal to
                        #examples.
                        For DataParallel models, each batch is split among the
                        available devices, so evaluations on each available
                        device contain internal_batch_size / num_devices examples.
                        If internal_batch_size is None, then all evaluations are
                        processed in one batch.
                        Default: None
            return_convergence_delta (bool, optional): Indicates whether to return
                    convergence delta or not. If `return_convergence_delta`
                    is set to True convergence delta will be returned in
                    a tuple following attributions.
                    Default: False
        Returns:
            **attributions** or 2-element tuple of **attributions**, **delta**:
            - **attributions** (*Tensor* or *tuple[Tensor, ...]*):
                    Integrated gradients with respect to each input feature.
                    attributions will always be the same size as the provided
                    inputs, with each value providing the attribution of the
                    corresponding input index.
                    If a single tensor is provided as inputs, a single tensor is
                    returned. If a tuple is provided for inputs, a tuple of
                    corresponding sized tensors is returned.
            - **delta** (*Tensor*, returned if return_convergence_delta=True):
                    The difference between the total approximated and true
                    integrated gradients. This is computed using the property
                    that the total sum of forward_func(inputs) -
                    forward_func(baselines) must equal the total sum of the
                    integrated gradient.
                    Delta is calculated per example, meaning that the number of
                    elements in returned delta tensor is equal to the number of
                    examples in inputs.

        Examples::

            >>> # ImageClassifier takes a single input tensor of images Nx3x32x32,
            >>> # and returns an Nx10 tensor of class probabilities.
            >>> net = ImageClassifier()
            >>> ig = IntegratedGradients(net)
            >>> input = torch.randn(2, 3, 32, 32, requires_grad=True)
            >>> # Computes integrated gradients for class 3.
            >>> attribution = ig.attribute(input, target=3)
        """
        # Keeps track whether original input is a tuple or not before
        # converting it into a tuple.
        is_inputs_tuple = _is_tuple(inputs)

        inputs, baselines = _format_input_baseline(inputs, baselines)

        _validate_input(inputs, baselines, n_steps, method)

        if internal_batch_size is not None:
            num_examples = inputs[0].shape[0]
            attributions = _batch_attribution(
                self,
                num_examples,
                internal_batch_size,
                n_steps,
                inputs=inputs,
                baselines=baselines,
                target=target,
                additional_forward_args=additional_forward_args,
                method=method,
            )
        else:
            attributions = self._attribute(
                inputs=inputs,
                baselines=baselines,
                target=target,
                additional_forward_args=additional_forward_args,
                n_steps=n_steps,
                method=method,
            )

        if return_convergence_delta:
            start_point, end_point = baselines, inputs
            # computes approximation error based on the completeness axiom
            delta = self.compute_convergence_delta(
                attributions,
                start_point,
                end_point,
                additional_forward_args=additional_forward_args,
                target=target,
            )
            return _format_output(is_inputs_tuple, attributions), delta
        return _format_output(is_inputs_tuple, attributions)

    def _attribute(
        self,
        inputs: Tuple[Tensor, ...],
        baselines: Tuple[Union[Tensor, int, float], ...],
        target: TargetType = None,
        additional_forward_args: Any = None,
        n_steps: int = 50,
        method: str = "gausslegendre",
        step_sizes_and_alphas: Union[None, Tuple[List[float], List[float]]] = None,
    ) -> Tuple[Tensor, ...]:
        if step_sizes_and_alphas is None:
            # retrieve step size and scaling factor for specified
            # approximation method
            step_sizes_func, alphas_func = approximation_parameters(method)
            step_sizes, alphas = step_sizes_func(n_steps), alphas_func(n_steps)
        else:
            step_sizes, alphas = step_sizes_and_alphas

        # scale features and compute gradients. (batch size is abbreviated as bsz)
        # scaled_features' dim -> (bsz * #steps x inputs[0].shape[1:], ...)
        scaled_features_tpl = tuple(
            torch.cat(
                [baseline + alpha * (input - baseline) for alpha in alphas], dim=0
            ).requires_grad_()
            for input, baseline in zip(inputs, baselines)
        )

        additional_forward_args = _format_additional_forward_args(
            additional_forward_args
        )
        # apply number of steps to additional forward args
        # currently, number of steps is applied only to additional forward arguments
        # that are nd-tensors. It is assumed that the first dimension is
        # the number of batches.
        # dim -> (bsz * #steps x additional_forward_args[0].shape[1:], ...)
        input_additional_args = (
            _expand_additional_forward_args(additional_forward_args, n_steps)
            if additional_forward_args is not None
            else None
        )
        expanded_target = _expand_target(target, n_steps)

        # grads: dim -> (bsz * #steps x inputs[0].shape[1:], ...)
        grads = self.gradient_func(
            forward_fn=self.forward_func,
            inputs=scaled_features_tpl,
            target_ind=expanded_target,
            additional_forward_args=input_additional_args,
        )

        # flattening grads so that we can multilpy it with step-size
        # calling contiguous to avoid `memory whole` problems
        scaled_grads = [
            grad.contiguous().view(n_steps, -1)
            * torch.tensor(step_sizes).view(n_steps, 1).to(grad.device)
            for grad in grads
        ]

        # aggregates across all steps for each tensor in the input tuple
        # total_grads has the same dimensionality as inputs
        total_grads = tuple(
            _reshape_and_sum(
                scaled_grad, n_steps, grad.shape[0] // n_steps, grad.shape[1:]
            )
            for (scaled_grad, grad) in zip(scaled_grads, grads)
        )

        # computes attribution for each tensor in input tuple
        # attributions has the same dimensionality as inputs
        if not self.multiplies_by_inputs:
            attributions = total_grads
        else:
            attributions = tuple(
                total_grad * (input - baseline)
                for total_grad, input, baseline in zip(total_grads, inputs, baselines)
            )
        return attributions

    def has_convergence_delta(self) -> bool:
        return True

    @property
    def multiplies_by_inputs(self):
        return self._multiply_by_inputs



#!/usr/bin/env python3

import torch
from captum.attr._utils.interpretable_input import TextTemplateInput, TextTokenInput
from parameterized import parameterized
from tests.helpers.basic import assertTensorAlmostEqual, BaseTest


class DummyTokenizer:
    def __init__(self, vocab_list):
        self.token_to_id = {v: i for i, v in enumerate(vocab_list)}
        self.id_to_token = vocab_list
        self.unk_idx = len(vocab_list) + 1

    def encode(self, text, **kwargs):
        return torch.tensor([self.convert_tokens_to_ids(text.split(" "))])

    def convert_ids_to_tokens(self, ids):
        return [
            (self.id_to_token[i] if i < len(self.id_to_token) else "[UNK]") for i in ids
        ]

    def convert_tokens_to_ids(self, tokens):
        return [
            (self.token_to_id[t] if t in self.token_to_id else self.unk_idx)
            for t in tokens
        ]


class TestTextTemplateInput(BaseTest):
    @parameterized.expand(
        [
            ("{} b {} {} e {}", ["a", "c", "d", "f"]),
            (
                "{arg1} b {arg2} {arg3} e {arg4}",
                {"arg1": "a", "arg2": "c", "arg3": "d", "arg4": "f"},
            ),
        ]
    )
    def test_input(self, template, values) -> None:
        tt_input = TextTemplateInput(template, values)

        expected_tensor = torch.tensor([[1.0] * 4])
        assertTensorAlmostEqual(self, tt_input.to_tensor(), expected_tensor)

        self.assertEqual(tt_input.to_model_input(), "a b c d e f")

        perturbed_tensor = torch.tensor([[1.0, 0.0, 1.0, 0.0]])
        self.assertEqual(tt_input.to_model_input(perturbed_tensor), "a b  d e ")

    @parameterized.expand(
        [
            ("{} b {} {} e {}", ["a", "c", "d", "f"], ["w", "x", "y", "z"]),
            (
                "{arg1} b {arg2} {arg3} e {arg4}",
                {"arg1": "a", "arg2": "c", "arg3": "d", "arg4": "f"},
                {"arg1": "w", "arg2": "x", "arg3": "y", "arg4": "z"},
            ),
        ]
    )
    def test_input_with_baselines(self, template, values, baselines) -> None:
        perturbed_tensor = torch.tensor([[1.0, 0.0, 1.0, 0.0]])

        # single instance baselines
        tt_input = TextTemplateInput(template, values, baselines=baselines)
        self.assertEqual(tt_input.to_model_input(perturbed_tensor), "a b x d e z")

    @parameterized.expand(
        [
            ("{} b {} {} e {}", ["a", "c", "d", "f"], [0, 1, 0, 1]),
            (
                "{arg1} b {arg2} {arg3} e {arg4}",
                {"arg1": "a", "arg2": "c", "arg3": "d", "arg4": "f"},
                {"arg1": 0, "arg2": 1, "arg3": 0, "arg4": 1},
            ),
        ]
    )
    def test_input_with_mask(self, template, values, mask) -> None:
        tt_input = TextTemplateInput(template, values, mask=mask)

        expected_tensor = torch.tensor([[1.0] * 2])
        assertTensorAlmostEqual(self, tt_input.to_tensor(), expected_tensor)

        self.assertEqual(tt_input.to_model_input(), "a b c d e f")

        perturbed_tensor = torch.tensor([[1.0, 0.0]])
        self.assertEqual(tt_input.to_model_input(perturbed_tensor), "a b  d e ")

    @parameterized.expand(
        [
            ("{} b {} {} e {}", ["a", "c", "d", "f"], [0, 1, 0, 1]),
            (
                "{arg1} b {arg2} {arg3} e {arg4}",
                {"arg1": "a", "arg2": "c", "arg3": "d", "arg4": "f"},
                {"arg1": 0, "arg2": 1, "arg3": 0, "arg4": 1},
            ),
        ]
    )
    def test_format_attr(self, template, values, mask) -> None:
        tt_input = TextTemplateInput(template, values, mask=mask)

        attr = torch.tensor([[0.1, 0.2]])

        assertTensorAlmostEqual(
            self, tt_input.format_attr(attr), torch.tensor([[0.1, 0.2, 0.1, 0.2]])
        )


class TestTextTokenInput(BaseTest):
    def test_input(self):
        tokenizer = DummyTokenizer(["a", "b", "c"])
        tt_input = TextTokenInput("a c d", tokenizer)

        expected_tensor = torch.tensor([[1.0] * 3])
        assertTensorAlmostEqual(self, tt_input.to_tensor(), expected_tensor)

        expected_model_inp = torch.tensor([[0, 2, tokenizer.unk_idx]])
        assertTensorAlmostEqual(self, tt_input.to_model_input(), expected_model_inp)

        perturbed_tensor = torch.tensor([[1.0, 0.0, 0.0]])
        expected_perturbed_inp = torch.tensor([[0, 0, 0]])
        assertTensorAlmostEqual(
            self, tt_input.to_model_input(perturbed_tensor), expected_perturbed_inp
        )

    def test_input_with_baselines(self):
        tokenizer = DummyTokenizer(["a", "b", "c"])

        # int baselines
        tt_input = TextTokenInput("a c d", tokenizer, baselines=1)

        perturbed_tensor = torch.tensor([[1.0, 0.0, 0.0]])
        expected_perturbed_inp = torch.tensor([[0, 1, 1]])
        assertTensorAlmostEqual(
            self, tt_input.to_model_input(perturbed_tensor), expected_perturbed_inp
        )

        # str baselines
        tt_input = TextTokenInput("a c d", tokenizer, baselines="b")
        assertTensorAlmostEqual(
            self, tt_input.to_model_input(perturbed_tensor), expected_perturbed_inp
        )

    def test_input_with_skip_tokens(self):
        tokenizer = DummyTokenizer(["a", "b", "c"])

        # int skip tokens
        tt_input = TextTokenInput("a c d", tokenizer, skip_tokens=[0])

        expected_tensor = torch.tensor([[1.0] * 2])
        assertTensorAlmostEqual(self, tt_input.to_tensor(), expected_tensor)

        expected_model_inp = torch.tensor([[0, 2, tokenizer.unk_idx]])
        assertTensorAlmostEqual(self, tt_input.to_model_input(), expected_model_inp)

        perturbed_tensor = torch.tensor([[0.0, 0.0]])
        expected_perturbed_inp = torch.tensor([[0, 0, 0]])
        assertTensorAlmostEqual(
            self, tt_input.to_model_input(perturbed_tensor), expected_perturbed_inp
        )

        # str skip tokens
        tt_input = TextTokenInput("a c d", tokenizer, skip_tokens=["a"])
        assertTensorAlmostEqual(self, tt_input.to_tensor(), expected_tensor)
        assertTensorAlmostEqual(self, tt_input.to_model_input(), expected_model_inp)
        assertTensorAlmostEqual(
            self, tt_input.to_model_input(perturbed_tensor), expected_perturbed_inp
        )



#!/usr/bin/env python3

import glob
import warnings
from abc import abstractmethod
from os.path import join
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Union,
)

import torch
from captum._utils.av import AV
from captum._utils.common import _get_module_from_name, _parse_version
from captum._utils.gradient import (
    _compute_jacobian_wrt_params,
    _compute_jacobian_wrt_params_with_sample_wise_trick,
)
from captum._utils.progress import NullProgress, progress
from captum.influence._core.influence import DataInfluence
from captum.influence._utils.common import (
    _check_loss_fn,
    _format_inputs_dataset,
    _get_k_most_influential_helper,
    _gradient_dot_product,
    _load_flexible_state_dict,
    _self_influence_by_batches_helper,
)
from captum.log import log_usage
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader, Dataset


r"""

Note: methods starting with "_" are protected, not private, and can be overridden in
child classes.  They are not part of the API.

Implements abstract DataInfluence class and provides implementation details for
influence computation based on the logic provided in TracIn paper
(https://arxiv.org/abs/2002.08484).

The TracIn paper proposes an idealized notion of influence which can be represented by
the total amount a training example reduces loss for a test example via a training
process such as stochastic gradient descent. As this idealized notion of influence is
impractical to compute, the TracIn paper proposes instead to compute an influence
score, which uses a first-order approximation for the change in loss for a test example
by a training example, which is accumulated across saved model checkpoints. This
influence score is accumulated via a summed dot-product of gradient vectors for the
scores/loss of a test and training example.
"""

"""
TODO: Support for checkpoint type. Currently only supports model parameters as saved
checkpoints. Can use enum or string.

Potential implementation from design doc:
checkpoint_type (Enum = [Parameters | Loss_Grad]): For performance,
                saved / loaded checkpoints can be either model parameters, or
                gradient of the loss function on an input w.r.t parameters.
"""


class KMostInfluentialResults(NamedTuple):
    """
    This namedtuple stores the results of using the `influence` method. This method
    is implemented by all subclasses of `TracInCPBase` to calculate
    proponents / opponents. The `indices` field stores the indices of the
    proponents / opponents for each example in the test dataset. For example, if
    finding opponents, `indices[i][j]` stores the index in the training data of the
    example with the `j`-th highest influence score on the `i`-th example in the test
    dataset. Similarly, the `influence_scores` field stores the actual influence
    scores, so that `influence_scores[i][j]` is the influence score of example
    `indices[i][j]` in the training data on example `i` of the test dataset.
    Please see `TracInCPBase.influence` for more details.
    """

    indices: Tensor
    influence_scores: Tensor


class TracInCPBase(DataInfluence):
    """
    To implement the `influence` method, classes inheriting from `TracInCPBase` will
    separately implement the private `_self_influence`, `_get_k_most_influential`,
    and `_influence` methods. The public `influence` method is a wrapper for these
    private methods.
    """

    def __init__(
        self,
        model: Module,
        train_dataset: Union[Dataset, DataLoader],
        checkpoints: Union[str, List[str], Iterator],
        checkpoints_load_func: Callable = _load_flexible_state_dict,
        loss_fn: Optional[Union[Module, Callable]] = None,
        batch_size: Union[int, None] = 1,
        test_loss_fn: Optional[Union[Module, Callable]] = None,
    ) -> None:
        r"""
        Args:

            model (torch.nn.Module): An instance of pytorch model. This model should
                    define all of its layers as attributes of the model.
            train_dataset (torch.utils.data.Dataset or torch.utils.data.DataLoader):
                    In the `influence` method, we compute the influence score of
                    training examples on examples in a test batch.
                    This argument represents the training dataset containing those
                    training examples. In order to compute those influence scores, we
                    will create a Pytorch DataLoader yielding batches of training
                    examples that is then used for processing. If this argument is
                    already a Pytorch Dataloader, that DataLoader can be directly
                    used for processing. If it is instead a Pytorch Dataset, we will
                    create a DataLoader using it, with batch size specified by
                    `batch_size`. For efficiency purposes, the batch size of the
                    DataLoader used for processing should be as large as possible, but
                    not too large, so that certain intermediate quantities created
                    from a batch still fit in memory. Therefore, if
                    `train_dataset` is a Dataset, `batch_size` should be large.
                    If `train_dataset` was already a DataLoader to begin with,
                    it should have been constructed to have a large batch size. It is
                    assumed that the Dataloader (regardless of whether it is created
                    from a Pytorch Dataset or not) yields tuples. For a `batch` that is
                    yielded, of length `L`, it is assumed that the forward function of
                    `model` accepts `L-1` arguments, and the last element of `batch` is
                    the label. In other words, `model(*batch[:-1])` gives the output of
                    `model`, and `batch[-1]` are the labels for the batch.
            checkpoints (str, list[str], or Iterator): Either the directory of the
                    path to store and retrieve model checkpoints, a list of
                    filepaths with checkpoints from which to load, or an iterator which
                    returns objects from which to load checkpoints.
            checkpoints_load_func (Callable, optional): The function to load a saved
                    checkpoint into a model to update its parameters, and get the
                    learning rate if it is saved. By default uses a utility to load a
                    model saved as a state dict.
                    Default: _load_flexible_state_dict
            loss_fn (Callable, optional): The loss function applied to model.
                    Default: None
            batch_size (int or None, optional): Batch size of the DataLoader created to
                    iterate through `train_dataset`, if it is a Dataset.
                    `batch_size` should be chosen as large as possible so that certain
                    intermediate quantities created from a batch still fit in memory.
                    Specific implementations of `TracInCPBase` will detail the size of
                    the intermediate quantities. `batch_size` must be an int if
                    `train_dataset` is a Dataset. If `train_dataset`
                    is a DataLoader, then `batch_size` is ignored as an argument.
                    Default: 1
            test_loss_fn (Callable, optional): In some cases, one may want to use a
                    separate loss functions for training examples, i.e. those in
                    `train_dataset`, and for test examples, i.e. those
                    represented by the `inputs` and `targets` arguments to the
                    `influence` method. For example, if one wants to calculate the
                    influence score of a training example on a test example's
                    prediction for a fixed class, `test_loss_fn` could map from the
                    logits for all classes to the logits for a fixed class.
                    `test_loss_fn` needs to satisfy the same constraints as `loss_fn`.
                    If not provided, the loss function for test examples is assumed to
                    be the same as the loss function for training examples, i.e.
                    `loss_fn`.
                    Default: None
        """

        self.model = model

        if isinstance(checkpoints, str):
            self.checkpoints = AV.sort_files(glob.glob(join(checkpoints, "*")))
        elif isinstance(checkpoints, List) and isinstance(checkpoints[0], str):
            self.checkpoints = AV.sort_files(checkpoints)
        else:
            self.checkpoints = list(checkpoints)  # cast to avoid mypy error
        if isinstance(self.checkpoints, List):
            assert len(self.checkpoints) > 0, "No checkpoints saved!"

        self.checkpoints_load_func = checkpoints_load_func
        self.loss_fn = loss_fn
        # If test_loss_fn not provided, it's assumed to be same as loss_fn
        self.test_loss_fn = loss_fn if test_loss_fn is None else test_loss_fn
        self.batch_size = batch_size

        if not isinstance(train_dataset, DataLoader):
            assert isinstance(batch_size, int), (
                "since the `train_dataset` argument was a `Dataset`, "
                "`batch_size` must be an int."
            )
            self.train_dataloader = DataLoader(train_dataset, batch_size, shuffle=False)
        else:
            self.train_dataloader = train_dataset

        self.train_dataloader_len: Optional[int] = None
        try:

            # since we will calculate the number of batches in
            # `self.train_dataloader` whenever we use progress bar, calculate
            # it once in initialization, for re-use.
            self.train_dataloader_len = len(self.train_dataloader)
        except TypeError:
            warnings.warn(
                "Unable to determine the number of batches in training dataset "
                "`train_dataset`. Therefore, if showing the progress of computations, "
                "only the number of batches processed can be displayed, and not the "
                "percentage completion of the computation, nor any time estimates."
            )

    @abstractmethod
    def self_influence(
        self,
        inputs: Optional[Union[Tuple[Any, ...], DataLoader]] = None,
        show_progress: bool = False,
    ) -> Tensor:
        """
        If `inputs` is not specified calculates the self influence
        scores for the training dataset `train_dataset`. Otherwise, computes
        self influence scores for the examples in `inputs`,
        which is either a single batch or a Pytorch `DataLoader` that yields
        batches. Therefore, in this case, the computed self influence scores
        are *not* for the examples in training dataset `train_dataset`.
        Note that if `inputs` is a single batch, this
        will call `model` on that single batch, and if `inputs` yields
        batches, this will call `model` on each batch that is yielded. Therefore,
        please ensure that for both cases, the batch(es) that `model` is called
        with are not too large, so that there will not be an out-of-memory error.

        Args:
            inputs (tuple or DataLoader, optional): This specifies the
                    dataset for which self influence scores will be computed.
                    Either a single tuple of any, or a `DataLoader`, where each
                    batch yielded is a tuple of type any. In either case, the tuple
                    represents a single batch, where the last element is assumed to
                    be the labels for the batch. That is, `model(*batch[0:-1])`
                    produces the output for `model`, and `batch[-1]` are the labels,
                    if any. This is the same assumption made for each batch yielded
                    by training dataset `train_dataset`. Please see documentation for
                    the `train_dataset` argument to `TracInCP.__init__` for
                    more details on the assumed structure of a batch. If not provided
                    or `None`, self influence scores will be computed for training
                    dataset `train_dataset`, which yields batches satisfying the
                    above assumptions.
                    Default: None.
            show_progress (bool, optional): Computation of self influence scores can
                    take a long time if `inputs` represents many examples. If
                    `show_progress` is true, the progress of this computation will be
                    displayed. In more detail, this computation will iterate over all
                    checkpoints (provided as the `checkpoints` initialization argument)
                    in an outer loop, and iterate over all batches that
                    `inputs` represents in an inner loop. Therefore, the
                    total number of (checkpoint, batch) combinations that need to be
                    iterated over is
                    (# of checkpoints x # of batches that `inputs` represents).
                    If `show_progress` is True, the total progress of both the outer
                    iteration over checkpoints and the inner iteration over batches is
                    displayed. It will try to use tqdm if available for advanced
                    features (e.g. time estimation). Otherwise, it will fallback to a
                    simple output of progress.
                    Default: False

        Returns:
            self_influence_scores (Tensor): This is a 1D tensor containing the self
                    influence scores of all examples in `inputs`, regardless of
                    whether it represents a single batch or a `DataLoader` that yields
                    batches.
        """
        pass

    @abstractmethod
    def _get_k_most_influential(
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        k: int = 5,
        proponents: bool = True,
        show_progress: bool = False,
    ) -> KMostInfluentialResults:
        r"""
        Args:

            inputs (tuple): `inputs` is the test batch and is a tuple of
                    any, where the last element is assumed to be the labels for the
                    batch. That is, `model(*batch[0:-1])` produces the output for
                    `model`, and `batch[-1]` are the labels, if any. This is the same
                    assumption made for each batch yielded by training dataset
                    `train_dataset` - please see its documentation in `__init__` for
                    more details on the assumed structure of a batch.
            k (int, optional): The number of proponents or opponents to return per test
                    example.
                    Default: 5
            proponents (bool, optional): Whether seeking proponents (`proponents=True`)
                    or opponents (`proponents=False`)
                    Default: True
            show_progress (bool, optional): To compute the proponents (or opponents)
                    for the batch of examples, we perform computation for each batch in
                    training dataset `train_dataset`, If `show_progress` is
                    true, the progress of this computation will be displayed. In
                    particular, the number of batches for which the computation has
                    been performed will be displayed. It will try to use tqdm if
                    available for advanced features (e.g. time estimation). Otherwise,
                    it will fallback to a simple output of progress.
                    Default: False

        Returns:
            (indices, influence_scores) (namedtuple): `indices` is a torch.long Tensor
                    that contains the indices of the proponents (or opponents) for each
                    test example. Its dimension is `(inputs_batch_size, k)`, where
                    `inputs_batch_size` is the number of examples in `inputs`. For
                    example, if `proponents==True`, `indices[i][j]` is the index of the
                    example in training dataset `train_dataset` with the
                    k-th highest influence score for the j-th example in `inputs`.
                    `indices` is a `torch.long` tensor so that it can directly be used
                    to index other tensors. Each row of `influence_scores` contains the
                    influence scores for a different test example, in sorted order. In
                    particular, `influence_scores[i][j]` is the influence score of
                    example `indices[i][j]` in training dataset `train_dataset`
                    on example `i` in the test batch represented by `inputs`.
        """
        pass

    @abstractmethod
    def _influence(
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        show_progress: bool = False,
    ) -> Tensor:
        r"""
        Args:

            inputs (tuple): `inputs` is the test batch and is a tuple of
                    any, where the last element is assumed to be the labels for the
                    batch. That is, `model(*batch[0:-1])` produces the output for
                    `model`, and `batch[-1]` are the labels, if any. This is the same
                    assumption made for each batch yielded by training dataset
                    `train_dataset` - please see its documentation in `__init__` for
                    more details on the assumed structure of a batch.
            show_progress (bool, optional): To compute the influence of examples in
                    training dataset `train_dataset`, we compute the influence
                    of each batch. If `show_progress` is true, the progress of this
                    computation will be displayed. In particular, the number of batches
                    for which influence has been computed will be displayed. It will
                    try to use tqdm if available for advanced features (e.g. time
                    estimation). Otherwise, it will fallback to a simple output of
                    progress.
                    Default: False

        Returns:
            influence_scores (Tensor): Influence scores over the entire
                    training dataset `train_dataset`. Dimensionality is
                    (inputs_batch_size, src_dataset_size). For example:
                    influence_scores[i][j] = the influence score for the j-th training
                    example to the i-th example in the test batch.
        """
        pass

    @abstractmethod
    def influence(  # type: ignore[override]
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        k: Optional[int] = None,
        proponents: bool = True,
        unpack_inputs: bool = True,
        show_progress: bool = False,
    ) -> Union[Tensor, KMostInfluentialResults]:
        r"""
        This is the key method of this class, and can be run in 2 different modes,
        where the mode that is run depends on the arguments passed to this method:

        - influence score mode: This mode is used if `k` is None. This mode computes
          the influence score of every example in training dataset `train_dataset`
          on every example in the test batch represented by `inputs`.
        - k-most influential mode: This mode is used if `k` is not None, and an int.
          This mode computes the proponents or opponents of every example in the
          test batch represented by `inputs`. In particular, for each test example in
          the test batch, this mode computes its proponents (resp. opponents),
          which are the indices in the training dataset `train_dataset` of the
          training examples with the `k` highest (resp. lowest) influence scores on the
          test example. Proponents are computed if `proponents` is True. Otherwise,
          opponents are computed. For each test example, this method also returns the
          actual influence score of each proponent (resp. opponent) on the test
          example.

        Args:

            inputs (tuple): `inputs` is the test batch and is a tuple of
                    any, where the last element is assumed to be the labels for the
                    batch. That is, `model(*batch[0:-1])` produces the output for
                    `model`, and `batch[-1]` are the labels, if any. This is the same
                    assumption made for each batch yielded by training dataset
                    `train_dataset` - please see its documentation in `__init__` for
                    more details on the assumed structure of a batch.
            k (int, optional): If not provided or `None`, the influence score mode will
                    be run. Otherwise, the k-most influential mode will be run,
                    and `k` is the number of proponents / opponents to return per
                    example in the test batch.
                    Default: None
            proponents (bool, optional): Whether seeking proponents (`proponents=True`)
                    or opponents (`proponents=False`), if running in k-most influential
                    mode.
                    Default: True
            show_progress (bool, optional): For all modes, computation of results
                    requires "training dataset computations": computations for each
                    batch in the training dataset `train_dataset`, which may
                    take a long time. If `show_progress` is true, the progress of
                    "training dataset computations" will be displayed. In particular,
                    the number of batches for which computations have been performed
                    will be displayed. It will try to use tqdm if available for
                    advanced features (e.g. time estimation). Otherwise, it will
                    fallback to a simple output of progress.
                    Default: False

        Returns:
            The return value of this method depends on which mode is run.

            - influence score mode: if this mode is run (`k` is None), returns a 2D
              tensor `influence_scores` of shape `(input_size, train_dataset_size)`,
              where `input_size` is the number of examples in the test batch, and
              `train_dataset_size` is the number of examples in training dataset
              `train_dataset`. In other words, `influence_scores[i][j]` is the
              influence score of the `j`-th example in `train_dataset` on the `i`-th
              example in the test batch.
            - k-most influential mode: if this mode is run (`k` is an int), returns
              a namedtuple `(indices, influence_scores)`. `indices` is a 2D tensor of
              shape `(input_size, k)`, where `input_size` is the number of examples in
              the test batch. If computing proponents (resp. opponents),
              `indices[i][j]` is the index in training dataset `train_dataset` of the
              example with the `j`-th highest (resp. lowest) influence score (out of
              the examples in `train_dataset`) on the `i`-th example in the test
              dataset. `influence_scores` contains the corresponding influence scores.
              In particular, `influence_scores[i][j]` is the influence score of example
              `indices[i][j]` in `train_dataset` on example `i` in the test batch
              represented by `inputs`.
        """
        pass

    @classmethod
    def get_name(cls: Type["TracInCPBase"]) -> str:
        r"""
        Create readable class name.  Due to the nature of the names of `TracInCPBase`
        subclasses, simplies returns the class name.  For example, for a class called
        TracInCP, we return the string TracInCP.

        Returns:
            name (str): a readable class name
        """
        return cls.__name__


def _influence_route_to_helpers(
    influence_instance: TracInCPBase,
    inputs: Union[Tuple[Any, ...], DataLoader],
    k: Optional[int] = None,
    proponents: bool = True,
    **kwargs,
) -> Union[Tensor, KMostInfluentialResults]:
    """
    This is a helper function called by `TracInCP.influence` and
    `TracInCPFast.influence`. Those methods share a common logic in that they assume
    an instance of their respective classes implement 2 private methods
    (``_influence`, `_get_k_most_influential`), and the logic of
    which private method to call is common, as described in the documentation of the
    `influence` method. The arguments and return values of this function are the exact
    same as the `influence` method. Note that `influence_instance` refers to the
    instance for which the `influence` method was called.
    """
    if k is None:
        return influence_instance._influence(inputs, **kwargs)
    else:
        return influence_instance._get_k_most_influential(
            inputs,
            k,
            proponents,
            **kwargs,
        )


class TracInCP(TracInCPBase):
    def __init__(
        self,
        model: Module,
        train_dataset: Union[Dataset, DataLoader],
        checkpoints: Union[str, List[str], Iterator],
        checkpoints_load_func: Callable = _load_flexible_state_dict,
        layers: Optional[List[str]] = None,
        loss_fn: Optional[Union[Module, Callable]] = None,
        batch_size: Union[int, None] = 1,
        test_loss_fn: Optional[Union[Module, Callable]] = None,
        sample_wise_grads_per_batch: bool = False,
    ) -> None:
        r"""
        Args:

            model (torch.nn.Module): An instance of pytorch model. This model should
                    define all of its layers as attributes of the model.
            train_dataset (torch.utils.data.Dataset or torch.utils.data.DataLoader):
                    In the `influence` method, we compute the influence score of
                    training examples on examples in a test batch.
                    This argument represents the training dataset containing those
                    training examples. In order to compute those influence scores, we
                    will create a Pytorch DataLoader yielding batches of training
                    examples that is then used for processing. If this argument is
                    already a Pytorch Dataloader, that DataLoader can be directly
                    used for processing. If it is instead a Pytorch Dataset, we will
                    create a DataLoader using it, with batch size specified by
                    `batch_size`. For efficiency purposes, the batch size of the
                    DataLoader used for processing should be as large as possible, but
                    not too large, so that certain intermediate quantities created
                    from a batch still fit in memory. Therefore, if
                    `train_dataset` is a Dataset, `batch_size` should be large.
                    If `train_dataset` was already a DataLoader to begin with,
                    it should have been constructed to have a large batch size. It is
                    assumed that the Dataloader (regardless of whether it is created
                    from a Pytorch Dataset or not) yields tuples. For a `batch` that is
                    yielded, of length `L`, it is assumed that the forward function of
                    `model` accepts `L-1` arguments, and the last element of `batch` is
                    the label. In other words, `model(*batch[:-1])` gives the output of
                    `model`, and `batch[-1]` are the labels for the batch.
            checkpoints (str, list[str], or Iterator): Either the directory of the
                    path to store and retrieve model checkpoints, a list of
                    filepaths with checkpoints from which to load, or an iterator which
                    returns objects from which to load checkpoints.
            checkpoints_load_func (Callable, optional): The function to load a saved
                    checkpoint into a model to update its parameters, and get the
                    learning rate if it is saved. By default uses a utility to load a
                    model saved as a state dict.
                    Default: _load_flexible_state_dict
            layers (list[str] or None, optional): A list of layer names for which
                    gradients should be computed. If `layers` is None, gradients will
                    be computed for all layers. Otherwise, they will only be computed
                    for the layers specified in `layers`.
                    Default: None
            loss_fn (Callable, optional): The loss function applied to model. There
                    are two options for the return type of `loss_fn`. First, `loss_fn`
                    can be a "per-example" loss function - returns a 1D Tensor of
                    losses for each example in a batch. `nn.BCELoss(reduction="none")`
                    would be an "per-example" loss function. Second, `loss_fn` can be
                    a "reduction" loss function that reduces the per-example losses,
                    in a batch, and returns a single scalar Tensor. For this option,
                    the reduction must be the *sum* or the *mean* of the per-example
                    losses. For instance, `nn.BCELoss(reduction="sum")` is acceptable.
                    Note for the first option, the `sample_wise_grads_per_batch`
                    argument must be False, and for the second option,
                    `sample_wise_grads_per_batch` must be True.  Also note that for
                    the second option, if `loss_fn` has no "reduction" attribute,
                    the implementation assumes that the reduction is the *sum* of the
                    per-example losses.  If this is not the case, i.e. the reduction
                    is the *mean*, please set the "reduction" attribute of `loss_fn`
                    to "mean", i.e. `loss_fn.reduction = "mean"`.
                    Default: None
            batch_size (int or None, optional): Batch size of the DataLoader created to
                    iterate through `train_dataset`, if it is a Dataset.
                    `batch_size` should be chosen as large as possible so that certain
                    intermediate quantities created from a batch still fit in memory.
                    Specific implementations of `TracInCPBase` will detail the size of
                    the intermediate quantities. `batch_size` must be an int if
                    `train_dataset` is a Dataset. If `train_dataset`
                    is a DataLoader, then `batch_size` is ignored as an argument.
                    Default: 1
            test_loss_fn (Callable, optional): In some cases, one may want to use a
                    separate loss functions for training examples, i.e. those in
                    `train_dataset`, and for test examples, i.e. those
                    represented by the `inputs` and `targets` arguments to the
                    `influence` method. For example, if one wants to calculate the
                    influence score of a training example on a test example's
                    prediction for a fixed class, `test_loss_fn` could map from the
                    logits for all classes to the logits for a fixed class.
                    `test_loss_fn` needs satisfy the same constraints as `loss_fn`.
                    Thus, the same checks that we apply to `loss_fn` are also applied
                    to `test_loss_fn`, if the latter is provided. Note that the
                    constraints on both `loss_fn` and `test_loss_fn` both depend on
                    `sample_wise_grads_per_batch`. This means `loss_fn` and
                    `test_loss_fn` must either both be "per-example"  loss functions,
                    or both be "reduction" loss functions. If not provided, the loss
                    function for test examples is assumed to be the same as the loss
                    function for training examples, i.e. `loss_fn`.
                    Default: None
            sample_wise_grads_per_batch (bool, optional): PyTorch's native gradient
                    computations w.r.t. model parameters aggregates the results for a
                    batch and does not allow to access sample-wise gradients w.r.t.
                    model parameters. This forces us to iterate over each sample in
                    the batch if we want sample-wise gradients which is computationally
                    inefficient. We offer an implementation of batch-wise gradient
                    computations w.r.t. to model parameters which is computationally
                    more efficient. This implementation can be enabled by setting the
                    `sample_wise_grad_per_batch` argument to `True`, and should be
                    enabled if and only if the `loss_fn` argument is a "reduction" loss
                    function. For example, `nn.BCELoss(reduction="sum")` would be a
                    valid `loss_fn` if this implementation is enabled (see
                    documentation for `loss_fn` for more details). Note that our
                    current implementation enables batch-wise gradient computations
                    only for a limited number of PyTorch nn.Modules: Conv2D and Linear.
                    This list will be expanded in the near future.  Therefore, please
                    do not enable this implementation if gradients will be computed
                    for other kinds of layers.
                    Default: False
        """

        TracInCPBase.__init__(
            self,
            model,
            train_dataset,
            checkpoints,
            checkpoints_load_func,
            loss_fn,
            batch_size,
            test_loss_fn,
        )

        self.sample_wise_grads_per_batch = sample_wise_grads_per_batch

        # check `loss_fn`
        self.reduction_type = _check_loss_fn(
            self, loss_fn, "loss_fn", sample_wise_grads_per_batch
        )
        # check `test_loss_fn` if it was provided
        self.test_reduction_type = (
            self.reduction_type
            if test_loss_fn is None
            else _check_loss_fn(
                self, test_loss_fn, "test_loss_fn", sample_wise_grads_per_batch
            )
        )

        r"""
        TODO: Either restore model state after done (would have to place functionality
        within influence to restore after every influence call)? or make a copy so that
        changes to grad_requires aren't persistent after using TracIn.
        """
        self.layer_modules = None
        if layers is not None:
            assert isinstance(layers, List), "`layers` should be a list!"
            assert len(layers) > 0, "`layers` cannot be empty!"
            assert isinstance(
                layers[0], str
            ), "`layers` should contain str layer names."
            self.layer_modules = [
                _get_module_from_name(self.model, layer) for layer in layers
            ]
            for layer, layer_module in zip(layers, self.layer_modules):
                for name, param in layer_module.named_parameters():
                    if not param.requires_grad:
                        warnings.warn(
                            "Setting required grads for layer: {}, name: {}".format(
                                ".".join(layer), name
                            )
                        )
                        param.requires_grad = True

    @log_usage()
    def influence(  # type: ignore[override]
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        k: Optional[int] = None,
        proponents: bool = True,
        show_progress: bool = False,
        aggregate: bool = False,
    ) -> Union[Tensor, KMostInfluentialResults]:
        r"""
        This is the key method of this class, and can be run in 2 different modes,
        where the mode that is run depends on the arguments passed to this method.
        Below, we describe the 2 modes, when `aggregate` is false:

        - influence score mode: This mode is used if `k` is None. This mode computes
          the influence score of every example in training dataset `train_dataset`
          on every example in the test dataset represented by `inputs`.
        - k-most influential mode: This mode is used if `k` is not None, and an int.
          This mode computes the proponents or opponents of every example in the
          test dataset represented by `inputs`. In particular, for each test example in
          the test dataset, this mode computes its proponents (resp. opponents),
          which are the indices in the training dataset `train_dataset` of the
          training examples with the `k` highest (resp. lowest) influence scores on the
          test example. Proponents are computed if `proponents` is True. Otherwise,
          opponents are computed. For each test example, this method also returns the
          actual influence score of each proponent (resp. opponent) on the test
          example.

        When `aggregate` is True, this method computes "aggregate" influence scores,
        which for a given training example, is the *sum* of its influence scores over
        all examples in the test dataset. Below, we describe the 2 modes, when
        `aggregate` is True:

        - influence score mode: This mode is used if `k` is None. This mode computes
          the aggregate influence score of each example in training dataset
          `train_dataset` on the test dataset.
        - k-most influential mode: This mode is used if `k` is not None, and an int.
          This mode computes the "aggregate" proponents (resp. opponents), which are
          the indices in the training dataset `train_dataset` of the examples with the
          `k` highest (resp. lowest) aggregate influence scores on the test dataset.
          Proponents are computed if `proponents` is True. Otherwise, opponents are
          computed. This method also returns the actual aggregate influence scores
          of each proponent (resp. opponent) on the test dataset.

        Args:

            inputs (Tuple, or DataLoader): Either a single tuple of any, or a
                    `DataLoader`, where each batch yielded is a tuple of any. In
                    either case, the tuple represents a single batch, where the last
                    element is assumed to be the labels for the batch. That is,
                    `model(*batch[0:-1])` produces the output for `model`, and
                    and `batch[-1]` are the labels, if any. Here, `model` is model
                    provided in initialization. This is the same assumption made for
                    each batch yielded by training dataset `train_dataset`. Please see
                    documentation for the `train_dataset` argument to
                    `TracInCPFastRandProj.__init__` for more details on the assumed
                    structure of a batch.
            k (int, optional): If not provided or `None`, the influence score mode will
                    be run. Otherwise, the k-most influential mode will be run,
                    and `k` is the number of proponents / opponents to return per
                    example in the test batch.
                    Default: None
            proponents (bool, optional): Whether seeking proponents (`proponents=True`)
                    or opponents (`proponents=False`), if running in k-most influential
                    mode.
                    Default: True
            show_progress (bool, optional): For all modes, computation of results
                    requires "training dataset computations": computations for each
                    batch in the training dataset `train_dataset`, which may
                    take a long time. If `show_progress` is true, the progress of
                    "training dataset computations" will be displayed. In particular,
                    the number of batches for which computations have been performed
                    will be displayed. It will try to use tqdm if available for
                    advanced features (e.g. time estimation). Otherwise, it will
                    fallback to a simple output of progress.
                    Default: False
            aggregate (bool, optional): If true, return "aggregate" influence scores or
                    examples with the highest / lowest aggregate influence scores on
                    the test dataset, depending on the mode.

        Returns:
            The return value of this method depends on which mode is run, and whether
            `aggregate` is True of False.

            Below are the return values for the 2 modes, when `aggregate` is False:

            - influence score mode: if this mode is run (`k` is None), returns a 2D
              tensor `influence_scores` of shape `(input_size, train_dataset_size)`,
              where `input_size` is the number of examples in the test dataset, and
              `train_dataset_size` is the number of examples in training dataset
              `train_dataset`. In other words, `influence_scores[i][j]` is the
              influence score of the `j`-th example in `train_dataset` on the `i`-th
              example in the test dataset.
            - k-most influential mode: if this mode is run (`k` is an int), returns
              a namedtuple `(indices, influence_scores)`. `indices` is a 2D tensor of
              shape `(input_size, k)`, where `input_size` is the number of examples in
              the test dataset. If computing proponents (resp. opponents),
              `indices[i][j]` is the index in training dataset `train_dataset` of the
              example with the `j`-th highest (resp. lowest) influence score (out of
              the examples in `train_dataset`) on the `i`-th example in the test
              dataset. `influence_scores` contains the corresponding influence scores.
              In particular, `influence_scores[i][j]` is the influence score of example
              `indices[i][j]` in `train_dataset` on example `i` in the test dataset
              represented by `inputs`.

            Below are the return values for the 2 modes, when `aggregate` is True:

            - influence score mode: if this mode is run (`k` is None), returns a 2D
              tensor `influence_scores` of shape `(1, train_dataset_size)`, where
              `influence_scores[0][j] is the aggregate influence score of the `j`-th
              example in `train_dataset` on the test dataset.
            - k-most influential mode: if this mode is run (`k` is an int), returns a
              namedtuple `(indices, influence_scores)`. `indices` is a 2D tensor of
              shape `(1, k)`. If computing proponents (resp. opponents),
              `indices[0][j]` is the index in training dataset `train_dataset` of the
              example with the `j`-th highest (resp. lowest) aggregate influence score
              on the test dataset. `influence_scores` contains the corresponding
              aggregate influence scores. In particular, `influence_scores[0][j]` is
              the aggregate influence score of example `indices[0][j]` on the test
              dataset.
        """

        assert inputs is not None, (
            "`inputs` argument is required."
            "If you wish to calculate self influence scores,"
            " please use the `self_influence` method instead."
        )
        return _influence_route_to_helpers(
            self,
            inputs,
            k,
            proponents,
            show_progress=show_progress,
            aggregate=aggregate,
        )

    def _sum_jacobians(
        self,
        inputs: DataLoader,
        loss_fn: Optional[Union[Module, Callable]] = None,
        reduction_type: Optional[str] = None,
    ) -> Tuple[Tensor, ...]:
        """
        sums the jacobians of all examples in `inputs`. result is of the
        same format as layer_jacobians, but the batch dimension has size 1
        """
        inputs_iter = iter(inputs)

        inputs_batch = next(inputs_iter)

        def get_batch_contribution(inputs_batch):
            _input_jacobians = self._basic_computation_tracincp(
                inputs_batch[0:-1],
                inputs_batch[-1],
                loss_fn,
                reduction_type,
            )

            return tuple(
                torch.sum(jacobian, dim=0).unsqueeze(0) for jacobian in _input_jacobians
            )

        inputs_jacobians = get_batch_contribution(inputs_batch)

        for inputs_batch in inputs_iter:
            inputs_batch_jacobians = get_batch_contribution(inputs_batch)
            inputs_jacobians = tuple(
                [
                    inputs_jacobian + inputs_batch_jacobian
                    for (inputs_jacobian, inputs_batch_jacobian) in zip(
                        inputs_jacobians, inputs_batch_jacobians
                    )
                ]
            )

        return inputs_jacobians

    def _concat_jacobians(
        self,
        inputs: DataLoader,
        loss_fn: Optional[Union[Module, Callable]] = None,
        reduction_type: Optional[str] = None,
    ) -> Tuple[Tensor, ...]:
        all_inputs_batch_jacobians = [
            self._basic_computation_tracincp(
                inputs_batch[0:-1],
                inputs_batch[-1],
                loss_fn,
                reduction_type,
            )
            for inputs_batch in inputs
        ]

        return tuple(
            torch.cat(all_inputs_batch_jacobian, dim=0)
            for all_inputs_batch_jacobian in zip(*all_inputs_batch_jacobians)
        )

    @log_usage()
    def compute_intermediate_quantities(
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        aggregate: bool = False,
    ) -> Tensor:
        """
        Computes "embedding" vectors for all examples in a single batch, or a
        `Dataloader` that yields batches. These embedding vectors are constructed so
        that the influence score of a training example on a test example is simply the
        dot-product of their corresponding vectors. Allowing a `DataLoader`
        yielding batches to be passed in (as opposed to a single batch) gives the
        potential to improve efficiency, because we load each checkpoint only once in
        this method call. Thus if a `DataLoader` yielding batches is passed in, this
        reduces the total number of times each checkpoint is loaded for a dataset,
        compared to if a single batch is passed in. The reason we do not just increase
        the batch size is that for large models, large batches do not fit in memory.

        If `aggregate` is True, the *sum* of the vectors for all examples is returned,
        instead of the vectors for each example. This can be useful for computing the
        influence of a given training example on the total loss over a validation
        dataset, because due to properties of the dot-product, this influence is the
        dot-product of the training example's vector with the sum of the vectors in the
        validation dataset. Also, by doing the sum aggregation within this method as
        opposed to outside of it (by computing all vectors for the validation dataset,
        then taking the sum) allows memory usage to be reduced.

        Args:
            inputs (Tuple, or DataLoader): Either a single tuple of any, or a
                    `DataLoader`, where each batch yielded is a tuple of any. In
                    either case, the tuple represents a single batch, where the last
                    element is assumed to be the labels for the batch. That is,
                    `model(*batch[0:-1])` produces the output for `model`, and
                    and `batch[-1]` are the labels, if any. Here, `model` is model
                    provided in initialization. This is the same assumption made for
                    each batch yielded by training dataset `train_dataset`.
            aggregate (bool): Whether to return the sum of the vectors for all
                    examples, as opposed to vectors for each example.

        Returns:
            intermediate_quantities (Tensor): A tensor of dimension
                    (N, D * C). Here, N is the total number of examples in
                    `inputs` if `aggregate` is False, and 1, otherwise (so that
                    a 2D tensor is always returned). C is the number of checkpoints
                    passed as the `checkpoints` argument of `TracInCP.__init__`, and
                    each row represents the vector for an example. Regarding D: Let I
                    be the dimension of the output of the last fully-connected layer
                    times the dimension of the input of the last fully-connected layer.
                    If `self.projection_dim` is specified in initialization,
                    D = min(I * C, `self.projection_dim` * C). Otherwise, D = I * C.
                    In summary, if `self.projection_dim` is None, the dimension of each
                    vector will be determined by the size of the input and output of
                    the last fully-connected layer of `model`. Otherwise,
                    `self.projection_dim` must be an int, and random projection will be
                    performed to ensure that the vector is of dimension no more than
                    `self.projection_dim` * C. `self.projection_dim` corresponds to
                    the variable d in the top of page 15 of the TracIn paper:
                    https://arxiv.org/pdf/2002.08484.pdf.
        """
        # If `inputs` is not a `DataLoader`, turn it into one.
        inputs = _format_inputs_dataset(inputs)

        def get_checkpoint_contribution(checkpoint):
            assert (
                checkpoint is not None
            ), "None returned from `checkpoints`, cannot load."

            learning_rate = self.checkpoints_load_func(self.model, checkpoint)
            # get jacobians as tuple of tensors
            if aggregate:
                inputs_jacobians = self._sum_jacobians(
                    inputs, self.loss_fn, self.reduction_type
                )
            else:
                inputs_jacobians = self._concat_jacobians(
                    inputs, self.loss_fn, self.reduction_type
                )
            # flatten into single tensor
            return learning_rate * torch.cat(
                [
                    input_jacobian.flatten(start_dim=1)
                    for input_jacobian in inputs_jacobians
                ],
                dim=1,
            )

        return torch.cat(
            [
                get_checkpoint_contribution(checkpoint)
                for checkpoint in self.checkpoints
            ],
            dim=1,
        )

    def _influence_batch_tracincp(
        self,
        input_checkpoint_jacobians: List[Tuple[Any, ...]],
        train_batch: Tuple[Any, ...],
    ) -> Tensor:
        """
        computes influence scores for a single training batch.
        `input_checkpoint_jacobians` is the output of
        `_basic_computation_tracincp` applied to the test batch, for each checkpoint,
        computed by `_get_checkpoint_jacobians`.
        """

        def get_checkpoint_contribution(input_jacobians, checkpoint):

            assert (
                checkpoint is not None
            ), "None returned from `checkpoints`, cannot load."

            learning_rate = self.checkpoints_load_func(self.model, checkpoint)

            return (
                _gradient_dot_product(
                    input_jacobians,
                    self._basic_computation_tracincp(
                        train_batch[0:-1],
                        train_batch[-1],
                        self.loss_fn,
                        self.reduction_type,
                    ),
                )
                * learning_rate
            )

        batch_tracin_scores = get_checkpoint_contribution(
            input_checkpoint_jacobians[0], self.checkpoints[0]
        )

        for (input_jacobians, checkpoint) in zip(
            input_checkpoint_jacobians[1:], self.checkpoints[1:]
        ):
            batch_tracin_scores += get_checkpoint_contribution(
                input_jacobians, checkpoint
            )

        return batch_tracin_scores

    def _get_checkpoint_jacobians(
        self,
        inputs_dataset: DataLoader,
        aggregate: bool,
        loss_fn: Optional[Union[Module, Callable]] = None,
    ) -> List[Tuple[Tensor, ...]]:
        """
        computes the jacobians of all examples in `inputs_dataset`, for all
        checkpoints. if `aggregate` is True, the jacobians for examples are summed.
        returns a list where each element corresponds to a checkpoint. this logic is
        separated into a helper function because it is used by both `_influence` and
        `_get_k_most_influential`.
        """
        inputs_checkpoint_jacobians = []
        for checkpoint in self.checkpoints:
            self.checkpoints_load_func(self.model, checkpoint)
            if aggregate:
                inputs_checkpoint_jacobians.append(
                    self._sum_jacobians(inputs_dataset, loss_fn, self.reduction_type)
                )
            else:
                inputs_checkpoint_jacobians.append(
                    self._concat_jacobians(inputs_dataset, loss_fn, self.reduction_type)
                )
        return inputs_checkpoint_jacobians

    def _influence(
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        show_progress: bool = False,
        aggregate: bool = False,
    ) -> Tensor:
        r"""
        Computes the influence of examples in training dataset `train_dataset`
        on the examples in the test dataset represented by `inputs`.
        This implementation does not require knowing the number of training examples
        in advance. Instead, the number of training examples is inferred from the
        output of `self._basic_computation_tracincp`.

        Args:

            inputs_dataset (Tuple, or DataLoader): Either a single tuple of any, or a
                    `DataLoader`, where each batch yielded is a tuple of any. In
                    either case, the tuple represents a single batch, where the last
                    element is assumed to be the labels for the batch. That is,
                    `model(*batch[0:-1])` produces the output for `model`, and
                    and `batch[-1]` are the labels, if any. Here, `model` is model
                    provided in initialization. This is the same assumption made for
                    each batch yielded by training dataset `train_dataset`.
            show_progress (bool, optional): To compute the influence of examples in
                    training dataset `train_dataset`, we compute the influence
                    of each batch. If `show_progress` is true, the progress of this
                    computation will be displayed. In particular, the number of batches
                    for which influence has been computed will be displayed. It will
                    try to use tqdm if available for advanced features (e.g. time
                    estimation). Otherwise, it will fallback to a simple output of
                    progress.
                    Default: False
            aggregate (bool): Whether to return "aggregate" influence scores (see their
                    definition in `influence`).
                    Default: False

        Returns:
            influence_scores (Tensor): If `aggregate` is False, influence scores are
            returned as a 2D tensor whose shape is `(input_size, train_dataset_size)`,
            where `input_size` is the number of examples in the test dataset, and
            `train_dataset_size` is the number of examples in
            training dataset `train_dataset`. For example:
            `influence_scores[i][j]` is the influence score for the j-th training
            example to the i-th example in the test dataset. If `aggregate` is True,
            "aggregate" influence scores are returned as a 2D tensor whose shape is
            `(1, train_dataset_size)`. For example: `influence_scores[0][j]` is the
            aggregate influence score of the j-th training example on the test dataset.
        """
        # If `inputs` is not a `DataLoader`, turn it into one.
        inputs = _format_inputs_dataset(inputs)

        train_dataloader = self.train_dataloader

        if show_progress:
            train_dataloader = progress(
                train_dataloader,
                desc=(
                    f"Using {self.get_name()} to compute "
                    "influence for training batches"
                ),
                total=self.train_dataloader_len,
            )

        # create list of the outputs of `_basic_computation_tracincp`, for each
        # checkpoint, which are jacobians
        inputs_checkpoint_jacobians = self._get_checkpoint_jacobians(
            inputs, aggregate, self.test_loss_fn
        )

        return torch.cat(
            [
                self._influence_batch_tracincp(inputs_checkpoint_jacobians, batch)
                for batch in train_dataloader
            ],
            dim=1,
        )

    def _get_k_most_influential(
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        k: int = 5,
        proponents: bool = True,
        show_progress: bool = False,
        aggregate: bool = False,
    ) -> KMostInfluentialResults:
        r"""
        Args:

            inputs (Tuple, or DataLoader): Either a single tuple of any, or a
                    `DataLoader`, where each batch yielded is a tuple of any. In
                    either case, the tuple represents a single batch, where the last
                    element is assumed to be the labels for the batch. That is,
                    `model(*batch[0:-1])` produces the output for `model`, and
                    and `batch[-1]` are the labels, if any. Here, `model` is model
                    provided in initialization. This is the same assumption made for
                    each batch yielded by training dataset `train_dataset`.
            k (int, optional): The number of proponents or opponents to return per test
                    example.
                    Default: 5
            proponents (bool, optional): Whether seeking proponents (`proponents=True`)
                    or opponents (`proponents=False`)
                    Default: True
            show_progress (bool, optional): To compute the proponents (or opponents)
                    for the batch of examples, we perform computation for each batch in
                    training dataset `train_dataset`, If `show_progress` is
                    true, the progress of this computation will be displayed. In
                    particular, the number of batches for which the computation has
                    been performed will be displayed. It will try to use tqdm if
                    available for advanced features (e.g. time estimation). Otherwise,
                    it will fallback to a simple output of progress.
                    Default: False
            aggregate (bool): Whether to return with the highest / lowest "aggregate"
                    influence scores (see their definition in `influence`).

        Returns:
            (indices, influence_scores) (namedtuple): If `aggregate` is False,
                    `indices` is a 2D tensor of shape `(input_size, k)`, where
                    `input_size` is the number of examples in the test dataset. If
                    computing proponents (resp. opponents), `indices[i][j]` is the
                    index in training dataset `train_dataset` of the example with the
                    `j`-th highest (resp. lowest) influence score (out of the examples
                    in `train_dataset`) on the `i`-th example in the test dataset.
                    `influence_scores` contains the corresponding influence scores. In
                    particular, `influence_scores[i][j]` is the influence score of
                    example `indices[i][j]` in `train_dataset` on example `i` in the
                    test dataset represented by `inputs`. If `aggregate` is True,
                    `indices` is a 2D tensor of shape `(1, k)`. If computing proponents
                    (resp. opponents), `indices[0][j]` is the index in training dataset
                    `train_dataset` of the example with the `j`-th highest (resp.
                    lowest) aggregate influence score on the test dataset.
                    `influence_scores` contains the corresponding aggregate influence
                    scores. In particular, `influence_scores[0][j]` is the aggregate
                    influence score of example `indices[0][j]` on the test dataset.
        """
        # If `inputs` is not a `DataLoader`, turn it into one.
        inputs = _format_inputs_dataset(inputs)

        desc = (
            None
            if not show_progress
            else (
                (
                    f"Using {self.get_name()} to perform computation for "
                    f'getting {"proponents" if proponents else "opponents"}. '
                    "Processing training batches"
                )
            )
        )

        # create list of the outputs of `_basic_computation_tracincp`, for each
        # checkpoint, which are jacobians
        inputs_checkpoint_jacobians = self._get_checkpoint_jacobians(
            inputs, aggregate, self.test_loss_fn
        )

        return KMostInfluentialResults(
            *_get_k_most_influential_helper(
                self.train_dataloader,
                self._influence_batch_tracincp,
                inputs_checkpoint_jacobians,
                k,
                proponents,
                show_progress,
                desc,
            )
        )

    def _self_influence_by_checkpoints(
        self,
        inputs: Union[Tuple[Any, ...], DataLoader],
        show_progress: bool = False,
    ) -> Tensor:
        """
        Computes self influence scores for the examples in `inputs`, which is
        either a single batch or a Pytorch `DataLoader` that yields batches. Therefore,
        the computed self influence scores are *not* for the examples in training
        dataset `train_dataset` (unlike when computing self influence scores using the
        `influence` method). Note that if `inputs` is a single batch, this
        will call `model` on that single batch, and if `inputs` yields
        batches, this will call `model` on each batch that is yielded. Therefore,
        please ensure that for both cases, the batch(es) that `model` is called
        with are not too large, so that there will not be an out-of-memory error. This
        implementation performs an outer iteration over checkpoints, and an inner
        iteration over all batches that `inputs` represents. The pros of this
        implementation are that the checkpoints do not need to be loaded too many
        times.

        Args:
            batches (tuple or DataLoader): Either a single tuple of any, or a
                    `DataLoader`, where each batch yielded is a tuple of any. In
                    either case, the tuple represents a single batch, where the last
                    element is assumed to be the labels for the batch. That is,
                    `model(*batch[0:-1])` produces the output for `model`,
                    and `batch[-1]` are the labels, if any. This is the same
                    assumption made for each batch yielded by training dataset
                    `train_dataset`. Please see documentation for the
                    `train_dataset` argument to `TracInCP.__init__` for
                    more details on the assumed structure of a batch.
            show_progress (bool, optional): Computation of self influence scores can
                    take a long time if `inputs` represents many examples. If
                    `show_progress` is true, the progress of this computation will be
                    displayed. In more detail, this computation will iterate over all
                    checkpoints (provided as the `checkpoints` initialization argument)
                    in an outer loop, and iterate over all batches that
                    `inputs` represents in an inner loop. Thus if
                    `show_progress` is True, the progress of both the outer
                    iteration and the inner iterations will be displayed. To show
                    progress, it will try to use tqdm if available for advanced
                    features (e.g. time estimation). Otherwise, it will fallback to a
                    simple output of progress.
                    Default: False

        Returns:
            self_influence_scores (Tensor): This is a 1D tensor containing the self
                    influence scores of all examples in `inputs`, regardless of
                    whether it represents a single batch or a `DataLoader` that yields
                    batches.
        """
        # If `inputs` is not a `DataLoader`, turn it into one.
        inputs = _format_inputs_dataset(inputs)

        # If `show_progress` is true, create an outer progress bar that keeps track of
        # how many checkpoints have been processed
        if show_progress:
            # Try to determine length of inner progress bar if possible, with a default
            # of `None`.
            inputs_len = None
            try:
                inputs_len = len(inputs)
            except TypeError:
                warnings.warn(
                    "Unable to determine the number of batches in `inputs`. "
                    "Therefore, if showing the progress of the computation of self "
                    "influence scores, only the number of batches processed can be "
                    "displayed, and not the percentage completion of the computation, "
                    "nor any time estimates."
                )

        def calculate_via_vector_norm(layer_jacobian):
            # Helper to efficiently calculate vector norm if pytorch version permits.
            return (
                torch.linalg.vector_norm(
                    layer_jacobian,
                    dim=list(range(1, len(layer_jacobian.shape))),
                )
                ** 2
            )

        def calculate_via_flatten(layer_jacobian):
            return torch.sum(layer_jacobian.flatten(start_dim=1) ** 2, dim=1)

        def get_checkpoint_contribution(checkpoint):
            # This function returns a 1D tensor representing the contribution to the
            # self influence score for the given checkpoint, for all batches in
            # `inputs`. The length of the 1D tensor is the total number of
            # examples in `inputs`.
            assert (
                checkpoint is not None
            ), "None returned from `checkpoints`, cannot load."

            learning_rate = self.checkpoints_load_func(self.model, checkpoint)

            # This will store a list of the contribution of the self influence score
            # from each batch. Each element is a 1D tensor of length batch_size - the
            # batch size of each batch in `inputs` (they do not need to be all
            # the same)
            checkpoint_contribution = []

            _inputs = inputs
            # If `show_progress` is true, create an inner progress bar that keeps track
            # of how many batches have been processed for the current checkpoint
            if show_progress:
                _inputs = progress(
                    inputs,
                    desc=(
                        f"Using {self.get_name()} to compute self "
                        "influence. Processing batch"
                    ),
                    total=inputs_len,
                )

            for batch in _inputs:

                layer_jacobians = self._basic_computation_tracincp(
                    batch[0:-1],
                    batch[-1],
                    self.loss_fn,
                    self.reduction_type,
                )

                # Note that all variables in this function are for an entire batch.
                # Each `layer_jacobian` in `layer_jacobians` corresponds to a different
                # layer. `layer_jacobian` is the jacobian w.r.t to a given layer's
                # parameters. If the given layer's parameters are of shape *, then
                # `layer_jacobian` is of shape (batch_size, *). For each layer, we need
                # the squared jacobian for each example. So we square the jacobian and
                # sum over all dimensions except the 0-th (the batch dimension). We then
                # sum the contribution over all layers.  For Pytorch > 1.10 we use the
                # optimized torch.linalg.vector_norm as opposed to the explicit flatten.

                calculate_fn = calculate_via_flatten
                if _parse_version(torch.__version__) >= (1, 10, 0):
                    calculate_fn = calculate_via_vector_norm

                checkpoint_contribution.append(
                    torch.sum(
                        torch.stack(
                            [
                                calculate_fn(layer_jacobian)
                                for layer_jacobian in layer_jacobians
                            ],
                            dim=0,
                        ),
                        dim=0,
                    )
                    * learning_rate
                )

            # We concatenate the contributions from each batch into a single 1D tensor,
            # which represents the contributions for all batches in `inputs`

            return torch.cat(checkpoint_contribution, dim=0)

        if show_progress:
            checkpoints_progress = progress(
                desc=(
                    f"Using {self.get_name()} to compute self "
                    "influence. Processing checkpoint"
                ),
                total=len(self.checkpoints),
                mininterval=0.0,
            )
        else:
            checkpoints_progress = NullProgress()
        with checkpoints_progress:
            batches_self_tracin_scores = get_checkpoint_contribution(
                self.checkpoints[0]
            )
            checkpoints_progress.update()
            # The self influence score for all examples is the sum of contributions from
            # each checkpoint
            for checkpoint in self.checkpoints[1:]:
                batches_self_tracin_scores += get_checkpoint_contribution(checkpoint)
                checkpoints_progress.update()

        return batches_self_tracin_scores

    @log_usage()
    def self_influence(
        self,
        inputs: Optional[Union[Tuple[Any, ...], DataLoader]] = None,
        show_progress: bool = False,
        outer_loop_by_checkpoints: bool = False,
    ) -> Tensor:
        """
        Computes self influence scores for the examples in `inputs`, which is
        either a single batch or a Pytorch `DataLoader` that yields batches.
        If `inputs` is not specified or `None` calculates self influence
        score for the training dataset `train_dataset`. Note that if `inputs`
        is a single batch, this will call `model` on that single batch, and if
        `inputs` yields batches, this will call `model` on each batch that is
        yielded. Therefore, please ensure that for both cases, the batch(es) that
        `model` is called with are not too large, so that there will not be an
        out-of-memory error.
        Internally, this computation requires iterating both over the batches in
        `inputs`, as well as different model checkpoints. There are two ways
        this iteration can be done. If `outer_loop_by_checkpoints` is False, the outer
        iteration will be over batches, and the inner iteration will be over
        checkpoints. This has the pro that displaying the progress of the computation
        is more intuitive, involving displaying the number of batches for which self
        influence scores have been computed. If `outer_loop_by_checkpoints` is True,
        the outer iteration will be over checkpoints, and the inner iteration will be
        over batches. This has the pro that the checkpoints do not need to be loaded
        for each batch. For large models, loading checkpoints can be time-intensive.

        Args:
            inputs (tuple or DataLoader, optional): This specifies the
                    dataset for which self influence scores will be computed.
                    Either a single tuple of any, or a `DataLoader`, where each
                    batch yielded is a tuple of type any. In either case, the tuple
                    represents a single batch, where the last element is assumed to
                    be the labels for the batch. That is, `model(*batch[0:-1])`
                    produces the output for `model`, and `batch[-1]` are the labels,
                    if any. This is the same assumption made for each batch yielded
                    by training dataset `train_dataset`. Please see documentation for
                    the `train_dataset` argument to `TracInCP.__init__` for
                    more details on the assumed structure of a batch. If not provided
                    or `None`, self influence scores will be computed for training
                    dataset `train_dataset`, which yields batches satisfying the
                    above assumptions.
                    Default: None.
            show_progress (bool, optional): Computation of self influence scores can
                    take a long time if `inputs` represents many examples. If
                    `show_progress`is true, the progress of this computation will be
                    displayed. In more detail, if `outer_loop_by_checkpoints` is False,
                    this computation will iterate over all batches in an outer loop.
                    Thus if `show_progress` is True, the number of batches for which
                    self influence scores have been computed will be displayed. If
                    `outer_loop_by_checkpoints` is True, this computation will iterate
                    over all checkpoints (provided as the `checkpoints` initialization
                    argument) in an outer loop, and iterate over all batches that
                    `inputs` represents in an inner loop. Thus if
                    `show_progress` is True, the progress of both the outer
                    iteration and the inner iterations will be displayed. To show
                    progress, it will try to use tqdm if available for advanced
                    features (e.g. time estimation). Otherwise, it will fallback to a
                    simple output of progress.
                    Default: False
            outer_loop_by_checkpoints (bool, optional): If performing an outer
                    iteration over checkpoints; see method description for more
                    details.
                    Default: False
        """
        inputs = inputs if inputs is not None else self.train_dataloader
        if outer_loop_by_checkpoints:
            return self._self_influence_by_checkpoints(inputs, show_progress)
        return _self_influence_by_batches_helper(
            self._self_influence_by_checkpoints,
            self.get_name(),
            inputs,
            show_progress,
        )

    def _basic_computation_tracincp(
        self,
        inputs: Tuple[Any, ...],
        targets: Optional[Tensor] = None,
        loss_fn: Optional[Union[Module, Callable]] = None,
        reduction_type: Optional[str] = None,
    ) -> Tuple[Tensor, ...]:
        """
        For instances of TracInCP, computation of influence scores or self influence
        scores repeatedly calls this function for different checkpoints
        and batches. In particular, this function computes the jacobian of a loss
        function w.r.t. parameters in the `layers` initialization argument.

        Args:

            inputs (tuple[Any, ...]): A batch of examples, which could be a training
                    batch or test batch, depending which method is the caller. Does not
                    represent labels, which are passed as `targets`. The assumption is
                    that `model(*inputs)` produces the predictions for the batch.
            targets (tensor or None): If computing influence scores on a loss function,
                    these are the labels corresponding to the batch `inputs`.
                    Default: none
            loss_fn (Callable, optional): The loss function to use when computing the
                    jacobian.
            reduction_type (str, optional): The reduction type of `loss_fn`. This
                    argument is only used if `sample_wise_grads_per_batch` was true in
                    initialization.
        """
        if self.sample_wise_grads_per_batch:
            return _compute_jacobian_wrt_params_with_sample_wise_trick(
                self.model,
                inputs,
                targets,
                loss_fn,
                reduction_type,
                self.layer_modules,
            )
        return _compute_jacobian_wrt_params(
            self.model,
            inputs,
            targets,
            loss_fn,
            self.layer_modules,
        )



#!/usr/bin/env python3
from typing import Callable


class Perturbation:
    r"""
    All perturbation and attack algorithms extend this class. It enforces
    its child classes to extend and override core `perturb` method.
    """

    perturb: Callable
    r"""
    This method computes and returns the perturbed input for each input tensor.
    Deriving classes are responsible for implementing its logic accordingly.

    Specific adversarial attack algorithms that extend this class take relevant
    arguments.

    Args:

        inputs (Tensor or tuple[Tensor, ...]): Input for which adversarial attack
                    is computed. It can be provided as a single tensor or
                    a tuple of multiple tensors. If multiple input tensors
                    are provided, the batch sizes must be aligned across all
                    tensors.

    Returns:

        - **perturbed inputs** (*Tensor* or *tuple[Tensor, ...]*):
                    Perturbed input for each
                    input tensor. The perturbed inputs have the same shape and
                    dimensionality as the inputs.
                    If a single tensor is provided as inputs, a single tensor
                    is returned. If a tuple is provided for inputs, a tuple of
                    corresponding sized tensors is returned.
    """

    def __call__(self, *args, **kwargs):
        return self.perturb(*args, **kwargs)
